/*  Project: Open Source Gmini Tools
    file: CPUCore.cpp, abstraction of the TCC730 cores
 
    Copyright (C) 2004  The gmEmu Project <gmemu@donat.org>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//config
//#define BUFFER_OVERRUN_DETECTION


#include <iostream>
#include <iomanip>
#include <fstream>
using namespace std;
#include <assert.h>
#include <stdio.h>
#include "Emulator.h"
#include "decodeOP.h"
#include "formatOP.h"
#include "util.h"

#define BITS8  (0xFF)
#define BITS16 (0xFFFF)
#define BITS24 (0xFFFFFF)
#define BITS48 (0xFFFFFFFFFFFFLL)

const char CPUCore::errcodestr[ERRCODE_END][20] = { "SUCCESS", "ABSJUMP", "RELJUMPUP", "RELJUMPDN", "SUCCESS_END", "NOTIMPLEMENTED", "INVALID", "RESERVED", "OPSIZE", "MEMORY",
    "READONLY", "WRITEONLY", "KNOWNDATA",
   "OOPS", "BREAKPOINT"
};
const char CPUCore::stateREGfile[] = "stateREG.dat";
CPUCore::CPUCore( Emulator * gm ) {
   gmini = gm;
   steps = 0;
   R[0] = R[1] = R[2] = R[3] = R[4] = R[5] = R[6] = R[7] = R[8] = R[9] = R[10] =
     R[11] = R[12] = R[13] = R[14] = R[15] = E[0] = E[1] = E[2] = E[3] = E[4] = E[5] = E[6] = E[7] = 0;
   PCC = SR = 0;
   for( int i = 0; i < SSR_END; i++ )
     SPSR[i] = 0;
   
   for ( int i=0; i < 4; i++) {
     RP[i] = 0;
     A[i] = 0;
     X[i] = 0;
     S[i] = 0;
     SD[i] = 0;
   }

   for ( int i=0; i < 2; i++) {
     MA[i]=0;
     RPD[i]=0;
     MC[i]=0;
   }
   ADDRESS=0x04D6;
   MSR[0]=MSR[1]=MSR[2]=0;
   EC3 = false;
   P =0;
}
void CPUCore::sigint(  ) {
   stop = true;
}

const int magic = 0xFADE;

void CPUCore::saveState(  ) {
   ofstream regs( stateREGfile, ios::out );
   if( !regs ) {
      console.output( "Could not open register state file: %s\n", stateREGfile );
      return;
   }
   //regs << hex;
   for( int i = 0; i < 16; ++i )
      regs << R[i] << endl;
   for( int i = 0; i < 8; ++i )
      regs << ( u16 ) E[i] << endl;
   regs << PCC << endl;
   regs << SR << endl;
   for( int i = 0; i < SSR_END; ++i )
      regs << SPSR[i] << endl;

   for( int i = 0; i<4; ++i )
     regs << X[i] << endl;
   for (int i = 0; i < 4; i++)
     regs << A[i] << endl;
   for (int i = 0; i < 4; i++)
     regs << S[i] << endl;

   regs << ADDRESS << endl;
   for (int i = 0; i < 3; i++)
     regs << MSR[i] << endl;
   for (int i = 0; i < 4; i++)
     regs << RP[i] << endl;
   for (int i = 0; i < 4; i++)
     regs << SD[i] << endl;
   regs << SD0E << endl;
   regs << SD3E << endl;
   for (int i = 0; i < 2; i++)
     regs << RPD[i] << endl;
   for (int i = 0; i < 2; i++)
     regs << MC[i] << endl;
   regs << P << endl;
   for (int i = 0; i < 2; i++)
     regs << MA[i] << endl;
   regs << EC3 << endl;
   regs << magic << endl;
   regs.close(  );
}
void CPUCore::restoreState(  ) {
   ifstream regs( stateREGfile, ios::in );
   if( !regs ) {
      console.output( "Could not open register state file: %s\n", stateREGfile );
      return;
   }
   regs.seekg(0, ios::beg);

   //regs.setf(ios::hex);

   u16 temp;
   for( int i = 0; i < 16; ++i ) {
      regs >> temp;
      R[i] = temp;
   }
   for( int i = 0; i < 8; ++i ) {
      regs >> temp;
      E[i] = temp;
   }
   regs >> PCC;
   regs >> SR;
   for( int i = 0; i < SSR_END; ++i ) {
      regs >> temp;
      SPSR[i] = temp;
   }

   for (int i = 0; i < 4; i++)
     regs >> X[i];
   for (int i = 0; i < 4; i++)
     regs >> A[i];
   for (int i = 0; i < 4; i++)
     regs >> S[i];
   regs >> ADDRESS;
   for (int i = 0; i < 3; i++)
     regs >> MSR[i];
   for (int i = 0; i < 4; i++)
     regs >> RP[i];
   for (int i = 0; i < 4; i++)
     regs >> SD[i];
   regs >> SD0E;
   regs >> SD3E;
   for (int i = 0; i < 2; i++)
     regs >> RPD[i];
   for (int i = 0; i < 2; i++)
     regs >> MC[i];
   regs >> P;
   for (int i = 0; i < 2; i++)
     regs >> MA[i];

   regs >> EC3;
   
   if (regs.eof()) {
     console.output("Bug in save/restore regs early eof\n");
     dumpregs();
     exit(-1);
   }
   regs >> temp;
   regs.close(  );

   if (temp != magic) {
     console.output("Bug in save/restore regs: magic = %x\n", temp);
     dumpregs();
     exit(-1);
   }
}
void CPUCore::locatefunctions(  ) {
   u8 sigDebug[13] = { 0x10, 0x7F, 0xC8, 0x09, 0xAD, 0x1B, 0xAB, 0x41, 0xAC, 0x1A, 0xAA, 0x41, 0x9F };
   u8 sigLCD[11] = { 0xAE, 0xFE, 0xA6, 0xF6, 0xC8, 0x07, 0xAB, 0x1A, 0xAA, 0x41, 0x9F };
   debugbreak = gmini->memory.find( sigDebug, 13 );
   haveDebugBreak = debugbreak != 0xFFFFFFFF;
   debugbreak -= 10;
   if( haveDebugBreak )
      console.output( "found copyDebugString() @ 0x%x\n", debugbreak );
   lcdbreak = gmini->memory.find( sigLCD, 11 );
   haveLCDBreak = lcdbreak != 0xFFFFFFFF;
   if( haveLCDBreak )
      console.output( "found printLCD() @ 0x%x\n", lcdbreak );
   emudebugbreak = gmini->symbols->addrof("_emu_debug");
   fprfuncbreak = gmini->symbols->addrof("_fprfunc");
}

void CPUCore::setR( int n, u16 val ) {
      assert( n >= 0 && n <= 15 );
      R[n] = val;
}

u32 CPUCore::getData( Location loc ) {
   u32 val;
   u22 sp;
   switch ( loc.kind ) {
      case Location::REG16:
         val = r( loc.index );
         break;
      case Location::REG32:
         val = ( r( loc.index + 1 ) << 16 ) + r( loc.index );
         break;
      case Location::REG22:
         val = a( loc.index );
         break;
      case Location::STACK16:
         sp = a( 15 );
         val = gmini->memory.dread16( sp + loc.index );
         break;
      case Location::STACK32:
         sp = a( 15 );
         val = ( gmini->memory.dread16( sp + loc.index - 2 ) << 16 ) + gmini->memory.dread16( sp + loc.index );
         break;
      default:
         assert( false );
   }
   return val;
}

void CPUCore::dump(  ) {
   dumpops( );
   dumpregs(  );
}

void CPUCore::disasm(u22 address) {
  char opstring[256];
  u22 end = address + 32;
  while (address < end) {
    address += printOp( address, opstring );
    console.output( opstring );
  }
}

void CPUCore::dumpBacktrace( ostream & o ) {
   if( callstack.empty(  ) ) {
      o << "Empty stack" << endl;
   }
   else {
      o << hex;
      for( unsigned i = 0; i < callstack.size(  ); i++ ) {
         StackFrame f = callstack[i];
         if( f.irq ) {
            o << "=== IRQ " << f.irqIndex << " ===" << endl;
         }
         o << i << ' ';
         u22 addr = f.called;
         o << f.from << " -> ";
         const Reference *ref = gmini->symbols->getReference( addr );
         if( ref == NULL )
            o << addr << endl;
         else {
            o << ref->getname(  ) << endl;
         }
      }
   }
}

char *condSel[] = {"Z",
		   "~Z",
		   "N",
		   "~Z ",
		   "C",
		   "~C",
		   "V",
		   "~V",
		   "GT",
		   "LE",
		   "VM0",
		   "VM1",
		   "VS",
		   "-",
		   "MV",
		   "T"};


void CPUCore::dumpregs(  ) {
   char dumpstring[65 * 10 + 1];

   u22 stackptr = a( 15 );
#define STACKVAL(X) (gmini->memory.dread16( stackptr + 2*((X)+1) ))
   
   sprintf( dumpstring,
            "Steps: %10d" "                      SP+ 2: %04x SP+22: %04x\n"
            "R0= %04x R8 = %04x E8 = %02x A8 = %06x"
            " SP+ 4: %04x SP+24: %04x\n"
            "R1= %04x R9 = %04x E9 = %02x A9 = %06x"
            " SP+ 6: %04x SP+26: %04x\n"
            "R2= %04x R10= %04x E10= %02x A10= %06x"
            " SP+ 8: %04x SP+28: %04x\n"
            "R3= %04x R11= %04x E11= %02x A11= %06x"
            " SP+10: %04x SP+30: %04x\n"
            "R4= %04x R12= %04x E12= %02x A12= %06x"
            " SP+12: %04x SP+32: %04x\n"
            "R5= %04x R13= %04x E13= %02x A13= %06x"
            " SP+14: %04x SP+34: %04x\n"
            "R6= %04x R14= %04x E14= %02x A14= %06x"
            " SP+16: %04x SP+36: %04x\n"
            "R7= %04x R15= %04x E15= %02x A15= %06x"
            " SP+18: %04x SP+38: %04x\n"
            "T:%d PM:%d Z1:%d Z0:%d V:%d TE:%d IE:%d FE:%d  SP+20: %04x SP+40: %04x\n",
            steps, STACKVAL(0), STACKVAL(10), R[0], R[8], E[0],
            ( E[0] << 16 ) | R[8], STACKVAL(1), STACKVAL(11), R[1], R[9],
            E[1], ( E[1] << 16 ) | R[9], STACKVAL(2), STACKVAL(12), R[2],
            R[10], E[2], ( E[2] << 16 ) | R[10], STACKVAL(3), STACKVAL(13),
            R[3], R[11], E[3], ( E[3] << 16 ) | R[11], STACKVAL(4),
            STACKVAL(14), R[4], R[12], E[4], ( E[4] << 16 ) | R[12],
            STACKVAL(5), STACKVAL(15), R[5], R[13], E[5],
            ( E[5] << 16 ) | R[13], STACKVAL(6), STACKVAL(16), R[6], R[14],
            E[6], ( E[6] << 16 ) | R[14], STACKVAL(7), STACKVAL(17), R[7],
            R[15], E[7], ( E[7] << 16 ) | R[15], STACKVAL(8), STACKVAL(18),
            flagT(  ), flagPM(  ), flagZ1(  ), flagZ0(  ), flagV(  ),
            flagTE(  ), flagIE(  ), flagFE(  ), STACKVAL(9), STACKVAL(19) );
   console.output( "%s", dumpstring );
   console.output("    DDSS  SP+42: %04x SP+44: %04x SP+46: %04x SP+48: %04x \n",
		  STACKVAL(20), STACKVAL(21), STACKVAL(22), STACKVAL(23)  );
   console.output("    1010\n");
   console.output("SD0=%04x RP0=%04x RPD0=%04x Y0=%06x A=%06x SA=%06x\n",SD[0],RP[0],RPD[0], X[0], A[0],S[0]); 
   console.output("SD1=%04x RP1=%04x RPD1=%04x Y1=%06x B=%06x SI=%06x\n",SD[1],RP[1],RPD[1],X[1],A[1], S[1]); 
   console.output("SD2=%04x RP2=%04x  MC0=%04x X0=%06x C=%06x SG=%06x\n",SD[2],RP[2],MC[0],X[2],A[2], S[2]); 
   console.output("SD3=%04x RP3=%04x  MC1=%04x X1=%06x D=%06x SR=%06x\n",SD[3], RP[3], MC[1], X[3], A[3], S[3]); 
   console.output("P=%012llx MA0=%x %012llx MA1=%x %012llx\n",P,
		  MAE(0), MA[0], 
		  MAE(1), MA[1]);
   console.output("T:%d C:%d Z:%d N:%d V:%d VS:%d OP:%d XSD:%d ME0:%d ME1:%d ME2:%d ME3:%d\n",
		  flagMAC(MAC_T), flagMAC(MAC_C), flagMAC(MAC_Z), flagMAC(MAC_N), flagMAC(MAC_V),
		  flagMAC(MAC_VS), flagMAC(OP), flagMAC(XSD), flagMAC(ME0), flagMAC(ME1), flagMAC(ME2), 
		  flagMAC(ME3));
   console.output("VM0:%d VM1:%d MV:%d OPM:%d USM:%d PSH1:%d NQ:%d OPMA:%d\n",
		  flagMAC1(VM0),flagMAC1(VM1),flagMAC1(MV),flagMAC1(OPM),flagMAC1(USM),flagMAC1(PSH1),
		  flagMAC1(NQ),flagMAC1(OPMA));

   console.output("EC0:%s EC1:%s EC2:%s EC3:%d BK0:%d BK1:%d BK2:%d BK3:%d\n", 
		  condSel[MSR[2]&0xF],
		  condSel[(MSR[2]>>4)&0xF],
		  condSel[(MSR[2]>>8)&0xF], EC3,
		  (MSR[2] >> BK0) & 1,
		  (MSR[2] >> BK1) & 1,
		  (MSR[2] >> BK2) & 1,
		  (MSR[2] >> BK3) & 1
		  );
}
void CPUCore::dumpops( ostream & o ) {
   char opstring[256];
   printOp( PCC, opstring );
   o << opstring;
}
void CPUCore::dumpops( ) {
   char opstring[256];
   printOp( PCC, opstring );
   console.output( opstring );
}

int CPUCore::printOp( u32 addr, char *opstring ) {
   u16 op1 = gmini->memory.cread( addr );
   u16 op2 = gmini->memory.cread( addr + 2 );
   int nnum = decodeOP( op1 );
   int size = opsize( nnum );
   int leng;
   if( size == 2 )
      leng = sprintf( opstring, " %06X: %04x      (%03d) %s\n", addr, op1, nnum,
                      formatOP( gmini->symbols, gmini->comments, nnum, op1, op2, addr + 2 ) );
   else
      leng =
        sprintf( opstring, " %06X: %04x %04x (%03d) %s\n", addr, op1, op2, nnum,
                 formatOP( gmini->symbols, gmini->comments, nnum, op1, op2, addr + 4 ) );
   const char *comment = gmini->comments->get( addr );
   if( comment != NULL ) {
      int j = leng;
      for( ; j < 45; ++j )
         opstring[j] = ' ';
      opstring[j++] = ';';
      leng += j + sprintf( opstring + j, " %s\n", comment );
   }
   return size;
}
void CPUCore::doTrace(  ) {
   // static u22 oldA8=0;
   // static bool traceEnable = false;
   // traceEnable = traceEnable || (PCC == 0x28300);
   // if (traceEnable) {
   // if (oldA8 != a(8)) {
   // trace << "A8 changed here!" << endl;
   // trace << "from " << hex << oldA8 << endl;
   // trace << "to " << hex << a(8) << endl;
   // dumpops(trace, 10);
   // }
   // oldA8 = a(8);
   // }
}

void CPUCore::go( u22 address ) {
   if( address == 0 )
      PCC = gmini->memory.cread16( 0 );
   else
      PCC = address;
}
void CPUCore::walk( u22 address ) {
   errcode err = SUCCESS;
   stop = false;
   while( !stop && ( PCC != address ) && ( err = step(  ) ) == SUCCESS ) {
      dumpregs(  );
      dumpops(  );
   }
   stopping( ( err != BREAKPOINT ) ? ( ( ( PCC == address ) ? REACHED : ERROR ) ) : BREAK, "walking" );
}
void CPUCore::jog( u22 address ) {
   errcode err = SUCCESS;
   stop = false;
   while( !stop && ( PCC != address ) && ( err = step(  ) ) == SUCCESS )
      dumpops(  );
   stopping( ( err != BREAKPOINT ) ? ( ( ( PCC == address ) ? REACHED : ERROR ) ) : BREAK, "jogging" );
}
void CPUCore::run( u22 address ) {
   errcode err = SUCCESS;
   stop = false;
   if( ( err = step(  ) ) == SUCCESS ) {
      while( !stop && ( PCC != address ) && ( err = step(  ) ) == SUCCESS ) ;
      stopping( ( err != BREAKPOINT ) ? ( ( ( PCC == address ) ? REACHED : ( stop ? SIGNAL : ERROR ) ) ) : BREAK,
                "running" );
   }
}
void CPUCore::run2ret(  ) {
   errcode err = SUCCESS;
   stop = false;
   unsigned depth = callstack.size(  );
   if( ( err = step(  ) ) == SUCCESS ) {
      while( !stop && ( callstack.size(  ) >= depth ) && ( err = step(  ) ) == SUCCESS ) ;
      stopping( ( err !=
                  BREAKPOINT ) ? ( ( ( callstack.size(  ) <= depth ) ? REACHED : ( stop ? SIGNAL : ERROR ) ) ) : BREAK,
                "running" );
   }
}
void CPUCore::stopping( reason_t why, char *what ) {
   console.output( "Stopped %s at 0x%x", what, PCC );
   switch ( why ) {
      case REACHED:
         console.output( ". The target address was reached.\n" );
         break;
      case ERROR:
         console.output( ". An error occurred.\n" );
         break;
      case BREAK:
         console.output( ". A breakpoint was reached.\n" );
         break;
      case SIGNAL:
         console.output( ". Received SIGINT\n" );
         break;
   }
   dumpops(  );
   dumpregs(  );
}

CPUCore::errcode CPUCore::step(  ) {
   if( gmini->BL.isBreakpoint( PCC, Breakpoint::X ) )
      return BREAKPOINT;
   u16 op = gmini->memory.cread( PCC );
   int mnnum = decodeOP( op );
#ifdef CALMMAC_DISABLED
   if( (mnnum > 164) && gmini->BL.breakPointEnabled ) {
      gmini->BL.breakPointEnabled = false;
      return BREAKPOINT;
   }
#endif
   int size = opsize( mnnum );
   errcode result = OOPS;
   u22 oldPC = PCC;
   PCC += size;
   if( size == 2 )
      result = doop( mnnum, op );
   else
      result = doop( mnnum, op, gmini->memory.cread( PCC - 2 ) ); // PCC already incremented
   if( result > SUCCESS_END )
      PCC = oldPC;
   else
      gmini->BL.breakPointEnabled = true;
      
   gmini->tracer.traceStep(  );
   switch ( result ) {
      case SUCCESS:
      case ABSJUMP:
      case RELJUMPUP:
      case RELJUMPDN:
   case SUCCESS_END:
         doOveride(  );
         return SUCCESS;
      case BREAKPOINT:
         doOveride(  );
         gmini->BL.Print( gmini->BL.getBreak(  ) );
         return BREAKPOINT;
      case NOTIMPLEMENTED:
      case INVALID:
      case RESERVED:
      case OPSIZE:
      case MEMORY:
      case READONLY:
      case WRITEONLY:
      case KNOWNDATA:
      case OOPS:
         console.output( "opcode returned: %s\n", errcodestr[result] );
         return OOPS;
      default:
         console.output( "opcode returned invalid error code\n" );
         return OOPS;
   }
}
void CPUCore::stepOver(  ) {
   errcode err = SUCCESS;
   stop = false;
   unsigned depth = callstack.size(  );
   if( ( err = step(  ) ) == SUCCESS ) {
      while( !stop && ( callstack.size(  ) > depth ) && ( err = step(  ) ) == SUCCESS ) ;
      stopping( ( err !=
                  BREAKPOINT ) ? ( ( ( callstack.size(  ) <= depth ) ? REACHED
                                     : ( stop ? SIGNAL : ERROR ) ) ) : BREAK, "running" );
   }
}

CPUCore::errcode CPUCore::delaySlot(  ) {
   u16 op = gmini->memory.cread( PCC );
   int mnnum = decodeOP( op );
   if( opsize( mnnum ) != 2 )
      return OPSIZE;
   return doop( mnnum, op );
}

void CPUCore::doOveride(  ) {
   char str[256];
   if( haveDebugBreak && PCC == debugbreak ) {
#ifdef LIMIT_IDE_DEBUG_OUTPUT
      // DIRTY HACK trim down debug output
      u22 address = a( 14 );
      if( address == 0x035fb6 || address == 0x036018 || address == 0x036020 ||
          address == 0x038652 || address == 0x037f66 || address == 0x038000 ||
          address == 0x03824a || address == 0x038298 || address == 0x0382a0 ||
          address == 0x0382c2 || address == 0x0382e4 || address == 0x038374 ||
          address == 0x0385e0 || address == 0x0385e8 || address == 0x03849e ||
          address == 0x038a82 || address == 0x038c7e )
         return;
#endif /* _LIMIT_IDE_DEBUG_OUTPUT_ */
      gmini->memory.getcstr( str, a( 10 ), 256 );
      console.output( "0x%06x: debugstring: 0x%06x: \"%s\"\n", a( 14 ), a( 10 ), str );
   }
#if 0
   if( haveLCDBreak && PCC == lcdbreak ) {

      gmini->memory.getcstr( str, a( 10 ), 256 );
      console.output( "0x%06x:   lcdstring: 0x%06x: \"%s\"\n", a( 14 ), a( 10 ), str );
   }
   if( PCC == 0x1EAE ) {
      gmini->memory.getdstr( str, a( 10 ), 256 );
      console.output( "0x%06x: coredbg: 0x%06x: \"%s\"\n", a( 14 ), a( 10 ), str );
   }
#endif
   if ( PCC == emudebugbreak) {
     u22 addr = gmini->memory.drawread32( a( 15 ) + 2);
     gmini->memory.getdstr( str, addr, 256 );
     console.output( "0x%06x: deb: 0x%06x: \"%s\" (SP=%x)\n", a( 14 ), addr, str, a(15));
     gmini->tracer.log("0x%06x: deb: 0x%06x: \"%s\" (SP=%x)\n", a( 14 ), addr, str, a(15));
   }

#if 0
   if (PCC == fprfuncbreak) {
     
     // render to uart.log, even if it should not go there.
     u8 c = gmini->memory.getdmem()[a( 15 ) + 7];
     gmini->uart << (char) c << flush;
     PCC = a(14); // return right away
   }
#endif
   
   if (steps % 16 == 0) {
     // in order not to slow the emulation too much,
     // costly emulation of non-core components is done only on certain cycles.


     // TIMERS
     timerIncrement( 0x02, MMIO::TACON, MMIO::TACNTH, MMIO::TADATAH );
     timerIncrement( 0x04, MMIO::TBCON, MMIO::TBCNTH, MMIO::TBDATAH );
     timerIncrement( 0x06, MMIO::TCCON, MMIO::TCCNTH, MMIO::TCDATAH );
     
     // UART
     u8 uartChar;
     u8 upend;
     if( ( steps % 32 == 0 ) &&
	 ( ( ( upend = gmini->io.rawread8( MMIO::UPEND ) ) & 1 ) == 0 ) && read( gmini->uartIn, &uartChar, 1 ) > 0 ) {
       gmini->io.rawwrite8( MMIO::RBR, uartChar );
       gmini->io.rawwrite8( MMIO::UPEND, upend | 1 );
       raiseIRQ( 0x0A );
     }

     //IIS
     //FIXME: should be done "real time" (instead i do a ad-hoc hack)
     if ((steps % 2048 == 0)&& (gmini->io.rawread8( MMIO::IISCON0 ) & 0x1 == 1)) {
       int iisptr = gmini->io.rawread8( MMIO::IISPTR0 );
       int data = gmini->io.rawread16(MMIO::IISBUF + iisptr*2);
       gmini->audio << hex << data << endl;
       iisptr++;
       iisptr &= 0x1F;
       if (iisptr == 0)
	 raiseIRQ( 13 );
       gmini->io.rawwrite8( MMIO::IISPTR0, iisptr );
     }
   }

   // check for interrupts
   if( flagIE(  ) ) {
      u32 imr = ( gmini->io.rawread16( MMIO::IMR1H ) << 16 ) + gmini->io.rawread16( MMIO::IMR0H );
      u32 irq = ( gmini->io.rawread16( MMIO::IRQ1H ) << 16 ) + gmini->io.rawread16( MMIO::IRQ0H );
      u32 ir = irq & imr;
      if( ir ) {
         // FIXME: priorities!
         int index = 0;
         while( ( ir & 1 ) == 0 ) {
            ir = ir >> 1;
            index++;
         }
         irq = irq & ~( 1 << index );
         gmini->io.rawwrite8( MMIO::IIR0H, index * 2 );
         gmini->io.rawwrite16( MMIO::IRQ1H, irq >> 16 );
         gmini->io.rawwrite16( MMIO::IRQ0H, irq );

         SPSR[SPCL_IRQ] = PCC;
         SPSR[SPCH_IRQ] = PCC >> 16;
         SPSR[SSR_IRQ] = SR;
         PCC = gmini->memory.cread16( 0x000004 );
         flagPM( true );
         flagIE( false );
         flagTE( false );

         StackFrame fr;
         fr.from = ( SPSR[SPCH_IRQ] << 16 ) | SPSR[SPCL_IRQ];
         fr.called = PCC;
         fr.irq = true;
         fr.irqIndex = index;
         callstack.push_back( fr );
      }
   }
}

void CPUCore::timerIncrement( int index, int CON, int CNT, int DATA ) {
  const int speed = 1;
   u8 tacon = 0;
   // do timer (this works only for specific case of the firmware settings)
   if( ( steps % 16 == 0 ) && ( ( tacon = gmini->io.rawread8( CON ) ) & 0x01 ) ) {
      // FIXME:
      // frequency depends on control register bits[32], and TAPRE
      gmini->io.rawwrite16( CNT, gmini->io.rawread16( CNT ) + speed );
      if( gmini->io.rawread16( CNT ) >= gmini->io.rawread16( DATA ) ) {
	// we should really increment by on and check for equality,
	// but for performance: increment by 'speed'; test for >=; adjust
	gmini->io.rawwrite16( CNT, gmini->io.rawread16( DATA ));
         if( tacon & 0x80 ) {
            gmini->io.rawwrite16( CNT, 0 );
            gmini->io.rawwrite8(CON, tacon & 0x7F);
         }
         raiseIRQ( index );
      }
   }
}

void CPUCore::raiseIRQ( int index ) {
   if( TRACE_IRQ )
      gmini->tracer.log( "IRQ raised %x", index );
   assert( index < 16 );
   gmini->io.rawwrite16( MMIO::IRQ0H, gmini->io.rawread16( MMIO::IRQ0H ) | ( 1 << index ) );
}

void CPUCore::overide( Overides what, u32 value ) {
   switch ( what ) {
      case ovFlagT:
         flagT( value );
         break;
      case ovFlagPM:
         flagPM( value );
         break;
      case ovFlagZ1:
         flagZ1( value );
         break;
      case ovFlagZ0:
         flagZ0( value );
         break;
      case ovFlagV:
         flagV( value );
         break;
      case ovFlagTE:
         flagTE( value );
         break;
      case ovFlagIE:
         flagIE( value );
         break;
      case ovFlagFE:
         flagFE( value );
         break;
      case ovR0:
      case ovR1:
      case ovR2:
      case ovR3:
      case ovR4:
      case ovR5:
      case ovR6:
      case ovR7:
      case ovR8:
      case ovR9:
      case ovR10:
      case ovR11:
      case ovR12:
      case ovR13:
      case ovR14:
      case ovR15:
         r( what - ovFlagFE, value );
         break;
      case ovE8:
      case ovE9:
      case ovE10:
      case ovE11:
      case ovE12:
      case ovE13:
      case ovE14:
      case ovE15:
         e( what - ovR15 + 8, value );
         break;
      case ovA8:
      case ovA9:
      case ovA10:
      case ovA11:
      case ovA12:
      case ovA13:
      case ovA14:
      case ovA15:
         a( what - ovE15 + 8, value );
   }
}
CPUCore::errcode CPUCore::branch( s22 off, bool save, bool delay ) {
   errcode result = SUCCESS;
   if( delay ) {
      result = delaySlot(  );
   }
   if( save ) {
      a( 14, PCC );
      call( PCC + off );
   }
   PCC += off;
   if( result == SUCCESS ) {
      if( off < 0 )
         result = RELJUMPUP;
      else if( off > 0 )
         result = RELJUMPDN;
   }
   return result;
}

void CPUCore::call( u22 addr ) {
   a( 14, PCC );
   char buffer[256];
   StackFrame f;
   f.called = addr;
   f.from = PCC;
   callstack.push_back( f );
   if( false ) {
      gmini->tracer.log( "Call interesting function" );
      const Reference *ref = gmini->symbols->getReference( addr );
      const vector < ParamDescr > params = ref->params;
      for( unsigned i = 0; i < params.size(  ); i++ ) {
         u32 v = getData( params[i].loc );
         const char *n = params[i].name.c_str(  );
         switch ( params[i].format ) {
            case ParamDescr::HEX:
               gmini->tracer.log( "param %s = %x", n, v );
               break;
            case ParamDescr::DEC:
               gmini->tracer.log( "param %s = %d", n, v );
               break;
            case ParamDescr::DSTR:
               gmini->memory.getdstr( buffer, v, 255 );
               gmini->tracer.log( "param %s = %x(%s)", n, v, buffer );
               break;
            case ParamDescr::CSTR:
               gmini->memory.getcstr( buffer, v, 255 );
               gmini->tracer.log( "param %s = %x(%s)", n, v, buffer );
               break;
            default:
               assert( false );
         }
      }
   }
}

CPUCore::errcode CPUCore::jump( u22 addr, bool save, bool delay ) {
   errcode result = SUCCESS;
   if( delay )
      result = delaySlot(  );
   if( save ) {
      a( 14, PCC );
      call( addr );
   }
   PCC = addr;
   if( result == SUCCESS )
      result = ABSJUMP;
   return result;
}

CPUCore::errcode CPUCore::subRet( u22 addr, bool delay ) {
   errcode res = jump( addr, false, delay );
#ifdef BUFFER_OVERRUN_DETECTION
   if( res == ABSJUMP ) {
      bool overrun = false;
      if( callstack.empty(  ) ) {
         overrun = true;
         console.output( "Buffer overrun detected: empty stack\n" );
      }
      else {
         StackFrame f = callstack.back(  );
         if( f.from != addr ) {
            overrun = true;
            console.output( "Buffer overrun detected: from=%x, addr=%x\n", f.from, addr );
         }

         callstack.pop_back(  );
      }

      if( overrun ) {
         res = OOPS;
      }
   }
#else
   if( res == ABSJUMP && !callstack.empty()) 
     callstack.pop_back(  );
#endif
   return res;
}

CPUCore::errcode CPUCore::doop( int mnnum, u16 op1, u16 op2 ) {
   ++steps;
   switch ( mnnum ) {
      case 0:
         return opADD2( op1 );
      case 1:
         return opSUB2( op1 );
      case 2:
         return opLD3( op1 );
      case 3:
         return opLDW1( op1 );
      case 4:
         return opLDW1( op1 );
      case 5:
         return opLDW2( op1 );
      case 6:
         return opLDW4( op1 );
      case 7:
         return opLDW2( op1 );
      case 8:
         return opLDW4( op1 );
      case 9:
         return opLDB1( op1 );
      case 10:
         return opLDB3( op1 );
      case 11:
         return opLDW5( op1 );
      case 12:
         return opLDW7( op1 );
      case 13:
         return opLDB1( op1 );
      case 14:
         return opLDB3( op1 );
      case 15:
         return opLDW5( op1 );
      case 16:
         return opLDW7( op1 );
      case 17:
         return opADD1( op1 );
      case 18:
         return opSUB1( op1 );
      case 19:
         return opADC1( op1 );
      case 20:
         return opSBC1( op1 );
      case 21:
         return opAND1( op1 );
      case 22:
         return opOR1( op1 );
      case 23:
         return opXOR1( op1 );
      case 24:
         return opTST1( op1 );
      case 25:
         return opCMP1( op1 );
      case 26:
         return opCMP1( op1 );
      case 27:
         return opCMP1( op1 );
      case 28:
         return opCMP1( op1 );
      case 29:
         return opCMP1( op1 );
      case 30:
         return opLD1( op1 );
      case 31:
         return opRR( op1 );
      case 32:
         return opRL( op1 );
      case 33:
         return opRRC( op1 );
      case 34:
         return opSRB( op1 );
      case 35:
         return opSR( op1 );
      case 36:
         return opSRA( op1 );
      case 37:
         return opJMP1( op1 );
      case 38:
         return opJMP1( op1 );
      case 39:
         return opJMP1( op1 );
      case 40:
         return opJMP1( op1 );
      case 41:
         return opSLB( op1 );
      case 42:
         return opDT( op1 );
      case 43:
         return opINCC( op1 );
      case 44:
         return opDECC( op1 );
      case 45:
         return opCOM( op1 );
      case 46:
         return opCOM( op1 );
      case 47:
         return opCOM( op1 );
      case 48:
         return opEXT( op1 );
      case 49:
         return opADD3( op1, op2 );
      case 50:
         return opADD5( op1, op2 );
      case 51:
         return opSUB4( op1, op2 );
      case 52:
         return opADC2( op1, op2 );
      case 53:
         return opSBC2( op1, op2 );
      case 54:
         return opAND3( op1, op2 );
      case 55:
         return opOR3( op1, op2 );
      case 56:
         return opXOR3( op1, op2 );
      case 57:
         return opTST3( op1, op2 );
      case 58:
         return opCMP2( op1, op2 );
      case 59:
         return opCMP2( op1, op2 );
      case 60:
         return opCMP2( op1, op2 );
      case 61:
         return opCMP2( op1, op2 );
      case 62:
         return opCMP2( op1, op2 );
      case 63:
         return opLD4( op1, op2 );
      case 64:
         return opReserved( op1 );
      case 65:
         return opCMPEQ2( op1 );
      case 66:
         return opAND2( op1 );
      case 67:
         return opOR2( op1 );
      case 68:
         return opXOR2( op1 );
      case 69:
         return opTST2( op1 );
      case 70:
         return opLDB4( op1 );
      case 71:
         return opLDB4( op1 );
      case 72:
         return opBIT( op1 );
      case 73:
         return opBIT( op1 );
      case 74:
         return opBIT( op1 );
      case 75:
         return opBIT( op1 );
      case 76:
         return opSYS( op1 );
      case 77:
         return opSWI( op1 );
      case 78:
         return opCLRSR( op1 );
      case 79:
         return opSETSR( op1 );
      case 80:
         return opTSTSR( op1 );
      case 81:
         return opNOP( op1 );
      case 82:
         return opBREAK( op1 );
      case 83:
         return opLDSR( op1 );
      case 84:
         return opLDSR( op1 );
      case 85:
         return opRET_FIQ( op1 );
      case 86:
         return opRET_IRQ( op1 );
      case 87:
         return opRET_SWI( op1 );
      case 88:
         return opRETD( op1 );
      case 89:
         return opLDSvR1( op1 );
      case 90:
         return opLDSvR1( op1 );
      case 91:
         return opLDSvR1( op1 );
      case 92:
         return opReserved( op1 );
      case 93:
         return opLDSvR1( op1 );
      case 94:
         return opLDSvR1( op1 );
      case 95:
         return opLDSvR1( op1 );
      case 96:
         return opReserved( op1 );
      case 97:
         return opReserved( op1 );
      case 98:
         return opLDSvR1( op1 );
      case 99:
         return opReserved( op1 );
      case 100:
         return opReserved( op1 );
      case 101:
         return opLDSvR2( op1 );
      case 102:
         return opLDSvR2( op1 );
      case 103:
         return opLDSvR2( op1 );
      case 104:
         return opReserved( op1 );
      case 105:
         return opLDSvR2( op1 );
      case 106:
         return opLDSvR2( op1 );
      case 107:
         return opLDSvR2( op1 );
      case 108:
         return opReserved( op1 );
      case 109:
         return opReserved( op1 );
      case 110:
         return opLDSvR2( op1 );
      case 111:
         return opReserved( op1 );
      case 112:
         return opReserved( op1 );
      case 113:
         return opReserved( op1 );
      case 114:
         return opReserved( op1 );
      case 115:
         return opLDPC( op1 );
      case 116:
         return opReserved( op1 );
      case 117:
         return opJMP2( op1, op2 );
      case 118:
         return opJMP2( op1, op2 );
      case 119:
         return opJMP2( op1, op2 );
      case 120:
         return opJMP2( op1, op2 );
      case 121:
         return opLDC( op1 );
      case 122:
         return opReserved( op1 );
      case 123:
         return opLDRE( op1 );
      case 124:
         return opLDRE( op1 );
      case 125:
         return opCMPEQ1( op1 );
      case 126:
         return opLD2( op1 );
      case 127:
         return opLDW3( op1, op2 );
      case 128:
         return opLDW3( op1, op2 );
      case 129:
         return opLDB2( op1, op2 );
      case 130:
         return opLDB2( op1, op2 );
      case 131:
         return opLDW6( op1, op2 );
      case 132:
         return opLDW6( op1, op2 );
      case 133:
         return opCMP3( op1 );
      case 134:
         return opADD6( op1 );
      case 135:
         return opSUB5( op1 );
      case 136:
         return opCMPEQ3( op1, op2 );
      case 137:
         return opLD5( op1, op2 );
      case 138:
         return opADD4( op1 );
      case 139:
         return opSUB3( op1 );
      case 140:
         return opMUL( op1 );
      case 141:
         return opMUL( op1 );
      case 142:
         return opMUL( op1 );
      case 143:
         return opMUL( op1 );
      case 144:
         return opPOP1( op1 );
      case 145:
         return opReserved( op1 );
      case 146:
         return opPOP2( op1 );
      case 147:
         return opPUSH1( op1 );
      case 148:
         return opReserved( op1 );
      case 149:
         return opPUSH2( op1 );
      case 150:
         return opBSRD( op1 );
      case 151:
         return opBRAEC( op1 );
      case 152:
         return opReserved( op1 );
      case 153:
         return opBRAEC( op1 );
      case 154:
         return opBNZD( op1 );
      case 155:
         return opReserved( op1 );
      case 156:
         return opBR( op1 );
      case 157:
         return opBR( op1 );
      case 158:
         return opBR( op1 );
      case 159:
         return opBR( op1 );
      case 160:
         return opBR( op1 );
      case 161:
         return opBR( op1 );
      case 164+0:
         return opECLD( op1 );
      case 164+1:
         return opELD5a( op1, op2 );
      case 164+2:
         return opEMOD0a( op1, op2 );
      case 164+3:
         return opELD5f( op1, op2 );
      case 164+4:
         return opERPN2( op1, op2 );
      case 169:
         return opELD1( op1, op2 );
      case 170:
         return opELD5( op1, op2 );
      case 164+7:
      case 164+8:
      case 164+9:
      case 164+10:
         return opEMUL0( op1 );
      case 164+11:
      case 164+12:
      case 164+16:
         return opEADDSUBLD1( op1 );
      case 164+13:
      case 164+14:
      case 164+15:
         return opEADDSUBLD2( op1 );
      case 164+17:
      case 164+18:
      case 164+19:
         return opEADDSUBLD3( op1 );
      case 164+20:
      case 164+21:
      case 164+22:
         return opEADDSUBLD4( op1 );
      case 187:
         return opReserved( op1 );
      case 164+24:
      case 164+25:
      case 164+26:
         return opEADDSUBLD5( op1 );
      case 164+27:
      case 164+28:
      case 164+29:
         return opEADDSUBLD6( op1 );
      case 164+30:
         return opELD5( op1 );
      case 164+31:
         return opReserved( op1 );
      case 164+32:
      case 164+33:
      case 164+34:
         return opEADDSUBLD7( op1 );
      case 164+35:
      case 164+36:
      case 164+37:
         return opEADDSUBLD8( op1 );
      case 164+38:
         return opELD5( op1 );
      case 203:
         return opReserved( op1 );
      case 164+40:
      case 164+41:
      case 164+42:
         return opEADDSUBLD9( op1 );
      case 164+43:
      case 164+44:
         return opEMINMAX2( op1 );
      case 164+45:
         return opELD5j( op1 );
      case 164+46:
         return opELD1a( op1 );
      case 211:
         return opELD5( op1 );
      case 212:
         return opELD5( op1 );
      case 164+49:
      case 164+50:
         return opELD5c( op1 );
      case 164+51:
      case 164+52:
      case 164+53:
      case 164+54:
         return opEMUL3( op1 );
      case 164+55:
         return opELD6( op1 );
      case 220:
         return opReserved( op1 );
      case 221:
         return opReserved( op1 );
      case 222:
      case 223:
      case 224:
      case 225:
         return opESFT( op1 );
      case 164+62:
         return opELD5e( op1 );
      case 227:
         return opENMSK( op1 );
      case 228:
         return opELD5b( op1 );
      case 229:
         return opELD5( op1 );
      case 230:
         return opELD5( op1 );
      case 164+67:
      case 164+68:
      case 164+69:
      case 164+70:
         return opESD( op1 );
      case 164+71:
      case 164+72:
         return opELD5d( op1 );
      case 164+73:
         return opEMOD0b(op1);
      case 164+74:
      case 164+75:
         return opEMOD0c(op1);
      case 164+76:
         return opELD5i( op1 );
      case 241:
         return opEMAD2( op1 );
      case 242:
         return opEMSB2( op1 );
      case 243:
         return opEMLD2( op1 );
      case 244:
         return opEMUL2( op1 );
      case 245:
      case 246:
         return opEMINMAX1( op1 );
      case 247:
      case 248:
         return opEMINMAX0( op1 );
      case 249:
         return opENOP( op1 );
      case 250:
         return opELD5( op1 );
      case 251:
         return opELD5( op1 );
      case 252:
         return opEMAD2( op1 );
      case 253:
         return opEMSB2( op1 );
      case 254:
         return opEMLD2( op1 );
      case 255:
         return opEMUL2( op1 );
      case 164+94:
      case 164+95:
      case 164+96:
      case 164+97:
         return opEMUL2( op1 );
      case 164+92:
      case 164+93:
      case 164+98:
      case 164+99:
         return opEMUL1( op1 );
      case 164+100:
         return opERPR( op1 );
      case 164+101:
      case 164+102:
      case 164+103:
         return opELD1b( op1 );
      case 164+104:
         return opEMOD1(op1);
      case 269:
         return opERPN1( op1 );
      case 270:
         return opERPS( op1 );
      case 271:
         return opERPD( op1 );
      case 164+108:
         return opEMOD2(op1);
      case 273:
         return opETST( op1 );
      case 274:
         return opES( op1 );
      case 164+111:
         return opELD5h( op1 );
      case 164+112:
         return opELD5h( op1 );
      case 164+113:
         return opELD5g( op1 );
      case 278:
         return opELD5( op1 );
      case 164+115:
      case 164+116:
      case 164+117:
      case 164+118:
         return opEBKESEC( op1 );
      default:
         return opInvalid( op1 );
   }
}
u16 CPUCore::sub2scomp( u16 val ) {
   val ^= 0xffff;
   return ++val;
}
CPUCore::errcode CPUCore::subADSBC( int n, u16 v2, bool c, bool m ) {
   u16 v1 = r( n );
   if( m )
      v2 = ( u16 ) ( sub2scomp( v2 ) - ( c ? 1 : 0 ) );
   u32 val = v1 + v2;
   if( c & !m ) {
      ++val;
   }
   if( m ) {
      flagT( !( val & 0x10000 ) && ( c || v2 ) );
      flagV( ( ( v1 & 0x8000 ) != ( v2 & 0x8000 ) ) &&
             ( ( u16 ) ( v1 & ( u16 ) 0x8000 ) != ( u16 ) ( val & 0x8000 ) ) );
   }
   else {
      flagT( val & 0x10000 );
      flagV( ( ( v1 & 0x8000 ) == ( v2 & 0x8000 ) ) &&
             ( ( u16 ) ( v1 & ( u16 ) 0x8000 ) != ( u16 ) ( val & 0x8000 ) ) );
   }
   r( n, ( u16 ) val );
   if( n == 6 )
      flagZ0( val == 0 );
   if( n == 7 )
      flagZ1( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::subSBC( int n, u16 v2, bool c ) {
   // return subADSBC(n,v2,c,true);
   u16 v1 = r( n );
   int val = v1 - v2 - 1 + ( c ? 1 : 0 );
   flagT( val >= 0 );
   flagV( ( ( v1 & 0x8000 ) != ( v2 & 0x8000 ) ) && ( ( u16 ) ( v1 & ( u16 ) 0x8000 ) != ( u16 ) ( val & 0x8000 ) ) );
   // FIXME: V flag ???
   r( n, ( u16 ) val );
   if( n == 6 )
      flagZ0( val == 0 );
   if( n == 7 )
      flagZ1( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::subSUB( int n, u16 v2 ) {
   // return subADSBC(n,v2,false,true);
   u16 v1 = r( n );
   int val = v1 - v2;
   flagT( val >= 0 );
   flagV( ( ( v1 & 0x8000 ) != ( v2 & 0x8000 ) ) && ( ( u16 ) ( v1 & ( u16 ) 0x8000 ) != ( u16 ) ( val & 0x8000 ) ) );
   // FIXME: V flag ???
   r( n, ( u16 ) val );
   if( n == 6 )
      flagZ0( val == 0 );
   if( n == 7 )
      flagZ1( val == 0 );
   return SUCCESS;
   // 0x2C3E0
}
CPUCore::errcode CPUCore::subAND( int n, u16 v2 ) {
   u16 v1 = r( n );
   u16 val = v1 & v2;
   flagT( val == 0 );
   r( n, val );
   if( n == 6 )
      flagZ0( val == 0 );
   if( n == 7 )
      flagZ1( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opADC1( u16 op1 ) {
   return subADSBC( ( op1 & 0xf00 ) >> 8, r( op1 & 0xf ), flagT(  ), false );
}
CPUCore::errcode CPUCore::opADC2( u16 op1, u16 op2 ) {
   return subADSBC( op1 & 0xf, op2, flagT(  ), false );
}
CPUCore::errcode CPUCore::opADD1( u16 op1 ) {
   return subADSBC( ( op1 & 0xf00 ) >> 8, r( op1 & 0xf ), false, false );
}
CPUCore::errcode CPUCore::opADD2( u16 op1 ) {
   return subADSBC( ( op1 & 0xf00 ) >> 8, ( u16 ) ( op1 & 0x7f ), false, false );
}
CPUCore::errcode CPUCore::opADD3( u16 op1, u16 op2 ) {
   return subADSBC( op1 & 0xf, op2, false, false );
}
CPUCore::errcode CPUCore::opADD4( u16 op1 ) {
   int n = ( ( op1 & 0x700 ) >> 8 ) + 8;
   int i = op1 & 0xf;
   a( n, a( n ) + r( i ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opADD5( u16 op1, u16 op2 ) {
   int n = ( op1 & 0x0007 ) + 8;
   a( n, a( n ) + op2 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opADD6( u16 op1 ) {
   int n = ( ( op1 & 0x700 ) >> 8 ) + 8;
   a( n, a( n ) + ( op1 & 0x1f ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opAND1( u16 op1 ) {
   return subAND( ( op1 & 0x0f00 ) >> 8, r( op1 & 0x000f ) );
}
CPUCore::errcode CPUCore::opAND2( u16 op1 ) {
   return subAND( 0, ( u16 ) ( op1 & 0x00ff ) );
}
CPUCore::errcode CPUCore::opAND3( u16 op1, u16 op2 ) {
   return subAND( op1 & 0x000f, op2 );
}
CPUCore::errcode CPUCore::opBIT( u16 op1 ) {
   u22 addr;
   u8 val, mask;
   addr = a( 8 ) + r( 1 );
   if( gmini->BL.isBreakpoint( addr, Breakpoint::R ) )
      return BREAKPOINT;
   if( ( ( op1 & 0x0018 ) >> 3 ) == 3 ) {
      if( gmini->BL.isBreakpoint( addr, Breakpoint::W ) )
         return BREAKPOINT;
   }
   val = gmini->memory.dread8( addr );
   mask = ( u8 ) ( 1 << ( op1 & 0x0007 ) );
   flagT( !( val & mask ) );
   switch ( ( op1 & 0x0018 ) >> 3 ) {
      case 0:
         val &= ( mask ^ ( u8 ) 0xff );
         break;
      case 1:
         val |= mask;
         break;
      case 2:
         val ^= mask;
         break;
      case 3:
         return SUCCESS;
   }
   return gmini->memory.dwrite8( addr, val ) ? SUCCESS : MEMORY;
}
CPUCore::errcode CPUCore::opBNZD( u16 op1 ) {
   int n = 7;
   if( !( op1 & 0x0100 ) )
      n = 6;
   u16 val = r( n );
   --val;
   r( n, val );
   s32 off = sext8to32( ( op1 & 0x007f ) << 1 );
   if( n == 6 && !flagZ0(  ) )
      flagZ0( val == 0 );
   else if( n == 7 && !flagZ1(  ) )
      flagZ1( val == 0 );
   else
      return SUCCESS;
   return branch( off, false, true );
}
CPUCore::errcode CPUCore::opBR( u16 op1 ) {
   bool delay = op1 & 0x0400;
   s32 off = sext11to32( ( op1 & 0x03ff ) << 1 );
   switch ( ( op1 & 0x1800 ) >> 11 ) {
      case 1:                  // BRA
         return branch( off, false, delay );
      case 2:                  // BRF
         if( !flagT(  ) )
            return branch( off, false, delay );
         break;
      case 3:                  // BRT
         if( flagT(  ) )
            return branch( off, false, delay );
         break;
      default:
         return INVALID;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opBRAEC( u16 op1 ) {
  bool delay = op1 & 0x0400;
  int ec = ( op1 & 0x0180 ) >> 7;
  bool cond;
  if (ec == 3)
    cond = EC3;
  else {
    int shift = ec * 4;
    int condSel = (MSR[2] >> shift) & 0xF;
// EC0 Selection
// 0000 = Z
// 0001 = ~Z
// 0010 = N
// 0011 = ~Z <- probably bug in datasheet
// 0100 = C
// 0101 = ~C
// 0110 = V
// 0111 = ~V
// 1000 = GT
// 1001 = LE
// 1010 = VM0
// 1011 = VM1
// 1100 = VS
// 1101 = reverved
// 1110 = MV
// 1111 = T

    switch(condSel) { 
      case 0x0:
	cond = flagMAC(MAC_Z);
	break;
      case 0x1:
	cond = !flagMAC(MAC_Z);
	break;
      case 0x2:
	cond = flagMAC(MAC_N);
	break;
      case 0x3:
	cond = !flagMAC(MAC_N);
	break;
      case 0x4:
	cond = flagMAC(MAC_C);
	break;
      case 0x5:
	cond = !flagMAC(MAC_C);
	break;
      case 0x6:
	cond = flagMAC(MAC_V);
	break;
      case 0x7:
	cond = !flagMAC(MAC_V);
	break;
      case 0x8:
	cond = !flagMAC(MAC_N) && !flagMAC(MAC_Z);
	break;
      case 0x9:
	cond = flagMAC(MAC_N) || flagMAC(MAC_Z);
	break;
      case 0xC:
	cond = flagMAC(MAC_VS);
	break;
      case 0xD:
	return RESERVED;
      case 0xF:
	cond = flagMAC(MAC_T);
	break;
    default:
      return NOTIMPLEMENTED;
    }
  }
  if(cond)
    return branch( sext8to32( ( op1 & 0x007f ) << 1 ), false, delay );
  return SUCCESS;
}
CPUCore::errcode CPUCore::opBREAK( u16 op1 ) {
   return SUCCESS;              // asserts nBRK signal for one cycle
}
CPUCore::errcode CPUCore::opBSRD( u16 op1 ) {
   return branch( sext11to32( ( op1 & 0x03ff ) << 1 ), true, true );
}
CPUCore::errcode CPUCore::opCLRSR( u16 op1 ) {
   switch ( op1 & 0x0007 ) {
      case 0:
         flagFE( false );
         break;
      case 1:
         flagIE( false );
         break;
      case 2:
         flagTE( false );
         break;
      case 3:
         flagV( false );
         break;
      case 4:
         flagZ0( false );
         break;
      case 5:
         flagZ1( false );
         break;
      case 6:
         flagPM( false );
         break;
      default:
         assert( false );
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opCMP1( u16 op1 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   int i = op1 & 0x000f;
   switch ( ( op1 & 0x0070 ) >> 4 ) {
      case 0:                  // GE
         flagT( ( s16 ) r( n ) >= ( s16 ) r( i ) );
         break;
      case 1:                  // GT
         flagT( ( s16 ) r( n ) > ( s16 ) r( i ) );
         break;
      case 2:                  // UGE
         flagT( r( n ) >= r( i ) );
         break;
      case 3:                  // UGT
         flagT( r( n ) > r( i ) );
         break;
      case 4:                  // EQ
         flagT( r( n ) == r( i ) );
         break;
      default:
         assert( false );
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opCMP2( u16 op1, u16 op2 ) {
   int n = op1 & 0x000f;
   switch ( ( op1 & 0x0700 ) >> 8 ) {
      case 0:                  // GE
         flagT( ( s16 ) r( n ) >= ( s16 ) op2 );
         break;
      case 1:                  // GT
         flagT( ( s16 ) r( n ) > ( s16 ) op2 );
         break;
      case 2:                  // UGE
         flagT( r( n ) >= op2 );
         break;
      case 3:                  // UGT
         flagT( r( n ) > op2 );
         break;
      case 4:                  // EQ
         flagT( r( n ) == op2 );
         break;
      default:
         assert( false );
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opCMP3( u16 op1 ) {
   flagT( ( s16 ) d( ( op1 & 0x0700 ) >> 8 ) >= ( s16 ) ( op1 & 0x003f ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opCMPEQ1( u16 op1 ) {
   flagT( a( ( ( op1 & 0x0700 ) >> 8 ) + 8 ) == a( ( op1 & 0x0007 ) + 8 ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opCMPEQ2( u16 op1 ) {
   flagT( d( ( op1 & 0x0700 ) >> 8 ) == ( op1 & 0x00ff ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opCMPEQ3( u16 op1, u16 op2 ) {
  flagT( a( ( ( op1 & 0x0700 ) >> 8 ) + 8 ) == ( u22 ) ( ( ( (u22)op1 & 0x003f ) << 16 ) | (u22)op2 ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opCOM( u16 op1 ) {
   int n = op1 & 0x000f;
   u16 val = r( n ) ^ ( u16 ) 0xffff;
   switch ( ( op1 & 0x0300 ) >> 8 ) {
      case 0:
         r( n, val );
         flagT( r( n ) == 0 );
         break;
      case 1:
         ++val;
         r( n, val );
         flagT( val == 0 );
         break;
      case 2:
         if( flagT(  ) )
            ++val;
         r( n, val );
         flagT( val == 0 );
         break;
      default:
         assert( false );
   }
   if( n == 6 )
      flagZ0( val == 0 );
   if( n == 7 )
      flagZ1( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opDECC( u16 op1 ) {
   u16 n = op1 & 0x000f;
   bool c = flagT(  );
   u32 v1 = r( n );
   u32 val = v1 - 1 + ( c ? 1 : 0 );
   flagT( val & 0x10000 );
   flagV( ( v1 == 0x8000 ) && !c );
   // FIXME: V flag ???
   r( n, ( u16 ) val );
   if( n == 6 )
      flagZ0( val == 0 );
   if( n == 7 )
      flagZ1( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opDT( u16 op1 ) {
   int n = op1 & 0x000f;
   u16 val = r( n );
   flagV( val == 0 );
   r( n, --val );
   flagT( val == 0 );
   if( n == 6 )
      flagZ0( val == 0 );
   if( n == 7 )
      flagZ1( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opEXT( u16 op1 ) {
   int n = op1 & 0x000f;
   u16 val = r( n );
   if( val & 0x80 )
      val |= 0xff00;
   else
      val &= 0x00ff;
   r( n, val );
   if( n == 6 )
      flagZ0( val == 0 );
   if( n == 7 )
      flagZ1( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opINCC( u16 op1 ) {
   return subADSBC( op1 & 0x000f, 0, flagT(  ), false );
}
CPUCore::errcode CPUCore::opJMP1( u16 op1 ) {
   int i = ( op1 & 0x0007 ) + 8;
   switch ( ( ( op1 & 0x0100 ) >> 7 ) | ( ( op1 & 0x0008 ) >> 3 ) ) {
      case 0:                  // JPF
         if( !flagT(  ) )
            return jump( a( i ) );
         break;
      case 1:                  // JPT
         if( flagT(  ) )
            return jump( a( i ) );
         break;
      case 2:                  // JMP
         if( i == 14 )
            return subRet( a( i ), false );
         else
            return jump( a( i ), false, false );
      case 3:                  // JSR
         return jump( a( i ), true );
      default:
         return INVALID;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opJMP2( u16 op1, u16 op2 ) {
   u22 imm = ( ( op1 & 0x003F ) << 16 ) | op2;
   switch ( ( op1 & 0x00c0 ) >> 6 ) {
      case 0:                  // JPF
         if( !flagT(  ) )
            return jump( imm );
         break;
      case 1:                  // JPT
         if( flagT(  ) )
            return jump( imm );
         break;
      case 2:                  // JMP
         return jump( imm );
      case 3:                  // JSR
         return jump( imm, true );
      default:
         return INVALID;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLD1( u16 op1 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   int i = ( op1 & 0x000f );
   u16 val = r( i );
   r( n, val );
   if( n == 6 )
      flagZ0( val == 0 );
   if( n == 7 )
      flagZ1( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLD2( u16 op1 ) {
   a( ( ( op1 & 0x0700 ) >> 8 ) + 8, a( ( op1 & 0x0007 ) + 8 ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLD3( u16 op1 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   u16 val = ( u16 ) ( op1 & 0x00ff );
   r( n, val );
   if( n == 6 )
      flagZ0( val == 0 );
   if( n == 7 )
      flagZ1( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLD4( u16 op1, u16 op2 ) {
   int n = op1 & 0x000f;
   r( n, op2 );
   if( n == 6 )
      flagZ0( op2 == 0 );
   if( n == 7 )
      flagZ1( op2 == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLD5( u16 op1, u16 op2 ) {
   a( ( ( op1 & 0x0700 ) >> 8 ) + 8, ( u6 ) ( op1 & 0x003f ), op2 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDRE( u16 op1 ) {
   int n = ( op1 & 0x0700 ) >> 8;
   int i = ( op1 & 0x0007 ) + 8;
   if( !( op1 & 0x0008 ) )
      d( n, ( u16 ) e( i ) );
   else
      e( i, ( u6 ) ( d( n ) & 0x003f ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDB1( u16 op1 ) {
   int n = ( op1 & 0x0700 ) >> 8;
   int i = ( ( op1 & 0x0070 ) >> 4 ) + 8;
   u22 disp = ( op1 & 0x000f );
   if( !( op1 & 0x1000 ) ) {
      if( gmini->BL.isBreakpoint( a( i ) + disp, Breakpoint::R ) )
         return BREAKPOINT;
      d( n, ( gmini->memory.dread8( a( i ) + disp ) & 0x00ff ) );
   }
   else {
      if( gmini->BL.isBreakpoint( a( i ) + disp, Breakpoint::W ) )
         return BREAKPOINT;
      return gmini->memory.dwrite8( a( i ) + disp, d( n ) ) ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDB2( u16 op1, u16 op2 ) {
   int n = ( op1 & 0x0700 ) >> 8;
   int i = ( op1 & 0x0007 ) + 8;
   u22 disp = op2;
   if( !( op1 & 0x0008 ) ) {
      if( gmini->BL.isBreakpoint( a( i ) + disp, Breakpoint::R ) )
         return BREAKPOINT;
      d( n, ( gmini->memory.dread8( a( i ) + disp ) & 0x00ff ) );
   }
   else {
      if( gmini->BL.isBreakpoint( a( i ) + disp, Breakpoint::W ) )
         return BREAKPOINT;
      return gmini->memory.dwrite8( a( i ) + disp, d( n ) ) ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDB3( u16 op1 ) {
   int n = ( op1 & 0x0700 ) >> 8;
   int i = ( ( op1 & 0x0070 ) >> 4 ) + 8;
   int j = ( op1 & 0x000f );
   if( !( op1 & 0x1000 ) ) {
      if( gmini->BL.isBreakpoint( a( i ) + r( j ), Breakpoint::R ) )
         return BREAKPOINT;
      d( n, ( gmini->memory.dread8( a( i ) + r( j ) ) & 0x00ff ) );
   }
   else {
      if( gmini->BL.isBreakpoint( a( i ) + r( j ), Breakpoint::W ) )
         return BREAKPOINT;
      return gmini->memory.dwrite8( a( i ) + r( j ), d( n ) ) ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDB4( u16 op1 ) {
   u22 disp = ( op1 & 0x00ff );
   if( !( op1 & 0x00100 ) ) {
      if( gmini->BL.isBreakpoint( a( 8 ) + disp, Breakpoint::R ) )
         return BREAKPOINT;
      r( 0, ( gmini->memory.dread8( a( 8 ) + disp ) & 0x00ff ) );
   }
   else {
      if( gmini->BL.isBreakpoint( a( 8 ) + disp, Breakpoint::W ) )
         return BREAKPOINT;
      return gmini->memory.dwrite8( a( 8 ) + disp, r( 0 ) ) ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDC( u16 op1 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   int i = ( op1 & 0x0007 ) + 8;
   u22 address = a( i ) & 0xFFFFFE;
   if( gmini->BL.isBreakpoint( address, Breakpoint::R ) )
      return BREAKPOINT;
   r( n, gmini->memory.cread16( address ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDPC( u16 op1 ) {
   int n = ( op1 & 0x0007 ) + 8;
   a( n, PCC & 0xFFFFFE );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDSvR1( u16 op1 ) {
   r( 0, SPSR[op1 & 0x000f] );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDSvR2( u16 op1 ) {
   SPSR[op1 & 0x000f] = r( 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDSR( u16 op1 ) {
   if( op1 & 0x0001 )
      r( 0, SR );
   else
      r( 0, SR );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW1( u16 op1 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   u22 disp = ( ( op1 & 0x00ff ) << 1 );
   if( !( op1 & 0x1000 ) ) {
      if( gmini->BL.isBreakpoint( ( a( 15 ) + disp ) & 0xFFFFFE, Breakpoint::R ) )
         return BREAKPOINT;
      r( n, gmini->memory.dread16( ( a( 15 ) + disp ) & 0xFFFFFE ) );
   }
   else {
      if( gmini->BL.isBreakpoint( ( a( 15 ) + disp ) & 0xFFFFFE, Breakpoint::W ) )
         return BREAKPOINT;
      return gmini->memory.dwrite16( ( a( 15 ) + disp ) & 0xFFFFFE, r( n ) ) ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW2( u16 op1 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   int m = ( ( op1 & 0x0070 ) >> 4 ) + 8;
   u22 disp = ( ( op1 & 0x000f ) << 1 );
   if( !( op1 & 0x1000 ) ) {
      if( gmini->BL.isBreakpoint( ( a( m ) + disp ) & 0xFFFFFE, Breakpoint::R ) )
         return BREAKPOINT;
      r( n, gmini->memory.dread16( ( a( m ) + disp ) & 0xFFFFFE ) );
   }
   else {
      if( gmini->BL.isBreakpoint( ( a( m ) + disp ) & 0xFFFFFE, Breakpoint::W ) )
         return BREAKPOINT;
      return gmini->memory.dwrite16( ( a( m ) + disp ) & 0xFFFFFE, r( n ) ) ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW3( u16 op1, u16 op2 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   int m = ( op1 & 0x0007 ) + 8;
   u22 disp = op2;
   if( !( op1 & 0x0008 ) ) {
      if( gmini->BL.isBreakpoint( ( a( m ) + disp ) & 0xFFFFFE, Breakpoint::R ) )
         return BREAKPOINT;
      r( n, gmini->memory.dread16( ( a( m ) + disp ) & 0xFFFFFE ) );
   }
   else {
      if( gmini->BL.isBreakpoint( ( a( m ) + disp ) & 0xFFFFFE, Breakpoint::W ) )
         return BREAKPOINT;
      return gmini->memory.dwrite16( ( a( m ) + disp ) & 0xFFFFFE, r( n ) ) ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW4( u16 op1 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   int m = ( ( op1 & 0x0070 ) >> 4 ) + 8;
   int i = ( op1 & 0x000f );
   if( !( op1 & 0x1000 ) ) {
      if( gmini->BL.isBreakpoint( ( a( m ) + r( i ) ) & 0xFFFFFE, Breakpoint::R ) )
         return BREAKPOINT;
      r( n, gmini->memory.dread16( ( a( m ) + r( i ) ) & 0xFFFFFE ) );
   }
   else {
      if( gmini->BL.isBreakpoint( ( a( m ) + r( i ) ) & 0xFFFFFE, Breakpoint::W ) )
         return BREAKPOINT;
      return gmini->memory.dwrite16( ( a( m ) + r( i ) ) & 0xFFFFFE, r( n ) ) ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW5( u16 op1 ) {
   int n = ( ( op1 & 0x0700 ) >> 8 ) + 8;
   int i = ( ( op1 & 0x0070 ) >> 4 ) + 8;
   u22 disp = ( ( op1 & 0x000f ) << 1 );
   if( !( op1 & 0x1000 ) ) {
      if( gmini->BL.isBreakpoint( ( a( i ) + disp ) & 0xFFFFFE, Breakpoint::R ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( ( a( i ) + disp + 2 ) & 0xFFFFFE, Breakpoint::R ) )
         return BREAKPOINT;
      a( n, gmini->memory.dread16( ( a( i ) + disp ) & 0xFFFFFE ),
         gmini->memory.dread16( ( a( i ) + disp + 2 ) & 0xFFFFFE ) );
   }
   else {
      bool ok = true;
      if( gmini->BL.isBreakpoint( ( a( i ) + disp ) & 0xFFFFFE, Breakpoint::W ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( ( a( i ) + disp + 2 ) & 0xFFFFFE, Breakpoint::W ) )
         return BREAKPOINT;
      ok &= gmini->memory.dwrite16( ( a( i ) + disp ) & 0xFFFFFE, e( n ) );
      ok &= gmini->memory.dwrite16( ( a( i ) + disp + 2 ) & 0xFFFFFE, r( n ) );
      return ok ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW6( u16 op1, u16 op2 ) {
   int n = ( ( op1 & 0x0700 ) >> 8 ) + 8;
   int i = ( op1 & 0x0007 ) + 8;
   u22 disp = op2;
   if( !( op1 & 0x0008 ) ) {
      if( gmini->BL.isBreakpoint( ( a( i ) + disp ) & 0xFFFFFE, Breakpoint::R ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( ( a( i ) + disp + 2 ) & 0xFFFFFE, Breakpoint::R ) )
         return BREAKPOINT;
      a( n, gmini->memory.dread16( ( a( i ) + disp ) & 0xFFFFFE ),
         gmini->memory.dread16( ( a( i ) + disp + 2 ) & 0xFFFFFE ) );
   }
   else {
      bool ok = true;
      if( gmini->BL.isBreakpoint( ( a( i ) + disp ) & 0xFFFFFE, Breakpoint::W ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( ( a( i ) + disp + 2 ) & 0xFFFFFE, Breakpoint::W ) )
         return BREAKPOINT;
      ok &= gmini->memory.dwrite16( ( a( i ) + disp ) & 0xFFFFFE, e( n ) );
      ok &= gmini->memory.dwrite16( ( a( i ) + disp + 2 ) & 0xFFFFFE, r( n ) );
      return ok ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW7( u16 op1 ) {
   int n = ( ( op1 & 0x0700 ) >> 8 ) + 8;
   int i = ( ( op1 & 0x0070 ) >> 4 ) + 8;
   int j = ( op1 & 0x000f );
   if( !( op1 & 0x1000 ) ) {
      if( gmini->BL.isBreakpoint( ( a( i ) + r( j ) ) & 0xFFFFFE, Breakpoint::R ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( ( a( i ) + r( j ) + 2 ) & 0xFFFFFE, Breakpoint::R ) )
         return BREAKPOINT;
      a( n, gmini->memory.dread16( ( a( i ) + r( j ) ) & 0xFFFFFE ),
         gmini->memory.dread16( ( a( i ) + r( j ) + 2 ) & 0xFFFFFE ) );
   }
   else {
      bool ok = true;
      if( gmini->BL.isBreakpoint( ( a( i ) + r( j ) ) & 0xFFFFFE, Breakpoint::W ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( ( a( i ) + r( j ) + 2 ) & 0xFFFFFE, Breakpoint::W ) )
         return BREAKPOINT;
      ok &= gmini->memory.dwrite16( ( a( i ) + r( j ) ) & 0xFFFFFE, e( n ) );
      ok &= gmini->memory.dwrite16( ( a( i ) + r( j ) + 2 ) & 0xFFFFFE, r( n ) );
      return ok ? SUCCESS : MEMORY;
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opMUL( u16 op1 ) {
   int n = ( ( op1 & 0x0700 ) >> 8 );
   int i = ( op1 & 0x0007 );
   u8 va = ( u8 ) ( d( n ) & 0xff );
   // bool sa = va & 0x80;
   u8 vb = ( u8 ) ( d( i ) & 0xff );
   // bool sb = vb & 0x80;
   u16 val;
   switch ( ( ( op1 & 0x0800 ) >> 10 ) | ( ( op1 & 0x0008 ) >> 3 ) ) {
      case 0:                  // UU
         val = ( u16 ) ( va * vb );
         break;
      case 1:                  // US
         val = ( u16 ) ( va * ( s8 ) vb );
         break;
      case 2:                  // SU
         val = ( u16 ) ( ( s8 ) va * vb );
         break;
      case 3:                  // SS
         val = ( u16 ) ( ( s8 ) va * ( s8 ) vb );
         break;
      default:
         return INVALID;
   }
   d( n, val );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opNOP( u16 op1 ) {
   return SUCCESS;
}
CPUCore::errcode CPUCore::opOR1( u16 op1 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   int i = op1 & 0x000f;
   u16 val = r( n ) | r( i );
   bool t = val == 0;
   flagT( t );
   r( n, val );
   if( n == 6 )
      flagZ0( t );
   if( n == 7 )
      flagZ1( t );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opOR2( u16 op1 ) {
   u16 val = r( 0 ) | ( u16 ) ( op1 & 0x00ff );
   flagT( val == 0 );
   r( 0, val );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opOR3( u16 op1, u16 op2 ) {
   int n = op1 & 0x000f;
   u16 val = r( op1 & 0x000f ) | op2;
   bool t = val == 0;
   flagT( t );
   r( n, val );
   if( n == 6 )
      flagZ0( t );
   if( n == 7 )
      flagZ1( t );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opPOP1( u16 op1 ) {
   int m = ( op1 & 0x0f00 ) >> 8;
   int n = ( op1 & 0x0007 ) | ( ( op1 & 0x0800 ) >> 8 );
   assert( m != 15 && n != 15 );
   u22 addr = a( 15 );
   if( m == n ) {
      if( gmini->BL.isBreakpoint( addr + 2, Breakpoint::R ) )
         return BREAKPOINT;
      r( n, gmini->memory.dread16( addr + 2 ) );
      a( 15, addr + 2 );
   }
   else {
      if( gmini->BL.isBreakpoint( addr + 2, Breakpoint::R ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( addr + 4, Breakpoint::R ) )
         return BREAKPOINT;
      r( n, gmini->memory.dread16( addr + 2 ) );
      r( m, gmini->memory.dread16( addr + 4 ) );
      a( 15, addr + 4 );
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opPOP2( u16 op1 ) {
   int m = ( ( op1 & 0x0700 ) >> 8 ) + 8;
   int n = ( op1 & 0x0007 ) + 8;
   assert( m != 15 && n != 15 );
   u22 addr = a( 15 );
   if( m == n ) {
      if( gmini->BL.isBreakpoint( addr + 2, Breakpoint::R ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( addr + 4, Breakpoint::R ) )
         return BREAKPOINT;
      e( n, ( u6 ) ( gmini->memory.dread16( addr + 2 ) & 0x003f ) );
      r( n, gmini->memory.dread16( addr + 4 ) );
      a( 15, addr + 4 );
   }
   else {
      if( gmini->BL.isBreakpoint( addr + 2, Breakpoint::R ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( addr + 4, Breakpoint::R ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( addr + 6, Breakpoint::R ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( addr + 8, Breakpoint::R ) )
         return BREAKPOINT;
      e( n, ( u6 ) ( gmini->memory.dread16( addr + 2 ) & 0x003f ) );
      r( n, gmini->memory.dread16( addr + 4 ) );
      e( m, ( u6 ) ( gmini->memory.dread16( addr + 6 ) & 0x003f ) );
      r( m, gmini->memory.dread16( addr + 8 ) );
      a( 15, addr + 8 );
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opPUSH1( u16 op1 ) {
   int m = ( op1 & 0x0f00 ) >> 8;
   int n = ( op1 & 0x0007 ) | ( ( op1 & 0x0800 ) >> 8 );
   assert( m != 15 && n != 15 );
   u22 addr = a( 15 );
   bool ok = true;
   if( m == n ) {
      if( gmini->BL.isBreakpoint( addr, Breakpoint::W ) )
         return BREAKPOINT;
      ok &= gmini->memory.dwrite16( addr, r( n ) );
      a( 15, addr - 2 );
   }
   else {
      if( gmini->BL.isBreakpoint( addr, Breakpoint::W ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( addr - 2, Breakpoint::W ) )
         return BREAKPOINT;
      ok &= gmini->memory.dwrite16( addr, r( n ) );
      ok &= gmini->memory.dwrite16( addr - 2, r( m ) );
      a( 15, addr - 4 );
   }
   return ok ? SUCCESS : MEMORY;
}
CPUCore::errcode CPUCore::opPUSH2( u16 op1 ) {
   int m = ( ( op1 & 0x0700 ) >> 8 ) + 8;
   int n = ( op1 & 0x0007 ) + 8;
   assert( m != 15 && n != 15 );
   u22 addr = a( 15 );
   bool ok = true;
   if( m == n ) {
      if( gmini->BL.isBreakpoint( addr, Breakpoint::W ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( addr - 2, Breakpoint::W ) )
         return BREAKPOINT;
      ok &= gmini->memory.dwrite16( addr, r( n ) );
      ok &= gmini->memory.dwrite16( addr - 2, e( n ) );
      a( 15, addr - 4 );
   }
   else {
      if( gmini->BL.isBreakpoint( addr, Breakpoint::W ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( addr - 2, Breakpoint::W ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( addr - 4, Breakpoint::W ) )
         return BREAKPOINT;
      if( gmini->BL.isBreakpoint( addr - 6, Breakpoint::W ) )
         return BREAKPOINT;
      ok &= gmini->memory.dwrite16( addr, r( n ) );
      ok &= gmini->memory.dwrite16( addr - 2, e( n ) );
      ok &= gmini->memory.dwrite16( addr - 4, r( m ) );
      ok &= gmini->memory.dwrite16( addr - 6, e( m ) );
      a( 15, addr - 8 );
   }
   return ok ? SUCCESS : MEMORY;
}
CPUCore::errcode CPUCore::opRETD( u16 op1 ) {
   return subRet( a( 14 ), true );
}
CPUCore::errcode CPUCore::opRET_FIQ( u16 op1 ) {
   SR = SPSR[SSR_FIQ];
   return jump( ( SPSR[SPCH_FIQ] << 16 ) | SPSR[SPCL_FIQ] );
}
CPUCore::errcode CPUCore::opRET_IRQ( u16 op1 ) {
   SR = SPSR[SSR_IRQ];
   u22 addr = ( SPSR[SPCH_IRQ] << 16 ) | SPSR[SPCL_IRQ];
   return subRet( addr, false );
}
CPUCore::errcode CPUCore::opRET_SWI( u16 op1 ) {
   SR = SPSR[SSR_SWI];
   return subRet( a( 14 ), false );
}
CPUCore::errcode CPUCore::opRL( u16 op1 ) {
  int n = op1 & 0xF;
  int v = r(n);
  bool bit = v & 0x8000;
  r(n, (v << 1) | (bit ? 1 : 0));
  flagT(bit);
  return SUCCESS;
}
CPUCore::errcode CPUCore::opRR( u16 op1 ) {
  int n = op1 & 0xF;
  int v = r(n);
  bool bit = v & 1;
  r(n, (v >> 1) | (bit ? 0x8000 : 0));
  flagT(bit);
  return SUCCESS;
}
CPUCore::errcode CPUCore::opRRC( u16 op1 ) {
   int n = op1 & 0x000f;
   u16 val = r( n );
   r( n, ( u16 ) ( ( val >> 1 ) | ( ( flagT(  )? 1 : 0 ) << 15 ) ) );
   flagT( val & 0x0001 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opSBC1( u16 op1 ) {
   return subSBC( ( op1 & 0x0f00 ) >> 8, r( op1 & 0x000f ), flagT(  ) );
}
CPUCore::errcode CPUCore::opSBC2( u16 op1, u16 op2 ) {
   return subSBC( op1 & 0x000f, op2, flagT(  ) );
}
CPUCore::errcode CPUCore::opSETSR( u16 op1 ) {
   switch ( op1 & 0x0007 ) {
      case 0:
         flagFE( true );
         break;
      case 1:
         flagIE( true );
         break;
      case 2:
         flagTE( true );
         break;
      case 3:
         flagV( true );
         break;
      case 4:
         flagZ0( true );
         break;
      case 5:
         flagZ1( true );
         break;
      case 6:
         flagPM( true );
         break;
      default:
         assert( false );
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opSLB( u16 op1 ) {
   int n = op1 & 0xf;
   u16 val = r( n );
   flagT( val & 0x100 );
   r( n, ( u16 ) ( val << 8 ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opSR( u16 op1 ) {
   int n = op1 & 0x000f;
   flagT( r( n ) & 0x0001 );
   r( n, ( u16 ) ( r( n ) >> 1 ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opSRA( u16 op1 ) {
   int n = op1 & 0xf;
   u16 val = r( n );
   r( n, ( u16 ) ( ( val >> 1 ) | ( val & 0x8000 ) ) );
   flagT( val & 1 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opSRB( u16 op1 ) {
   int n = op1 & 0xf;
   u16 val = r( n );
   flagT( val & 0x80 );
   r( n, ( u16 ) ( val >> 8 ) );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opSUB1( u16 op1 ) {
   return subSUB( ( op1 & 0xf00 ) >> 8, r( op1 & 0xf ) );
}
CPUCore::errcode CPUCore::opSUB2( u16 op1 ) {
   return subSUB( ( op1 & 0xf00 ) >> 8, ( u16 ) ( op1 & 0x7f ) );
}
CPUCore::errcode CPUCore::opSUB3( u16 op1 ) {
   int n = ( ( op1 & 0x0700 ) >> 8 ) + 8;
   int i = op1 & 0x000f;
   u22 val = ( u22 ) ( a( n ) - ( u22 ) r( i ) );
   a( n, val );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opSUB4( u16 op1, u16 op2 ) {
   int n = ( op1 & 0x0007 ) + 8;
   u22 val = ( u22 ) ( a( n ) - ( u22 ) op2 );
   a( n, val );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opSUB5( u16 op1 ) {
   int n = ( ( op1 & 0x0700 ) >> 8 ) + 8;
   u22 val = ( u22 ) ( a( n ) - ( u22 ) ( op1 & 0x001f ) );
   a( n, val );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opSWI( u16 op1 ) {
   int n = op1 & 0x3F;
   SPSR[SSR_SWI] = SR;
   flagIE( false );
   flagTE( false );
   u22 addr = gmini->memory.cread16( n * 4 );
   return jump( addr, true, false );
}
CPUCore::errcode CPUCore::opSYS( u16 op1 ) {
  int imm5 = op1 & 0x001f;
  console.output("SYS Interaction %02x.\n", imm5);
  return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opTST1( u16 op1 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   int i = op1 & 0x000f;
   u16 val = r( n ) & r( i );
   flagT( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opTST2( u16 op1 ) {
   u16 imm = op1 & ( u16 ) 0x00ff;
   u16 val = r( 0 ) & imm;
   flagT( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opTST3( u16 op1, u16 op2 ) {
   int n = op1 & 0x000f;
   u16 val = r( n ) & op2;
   flagT( val == 0 );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opTSTSR( u16 op1 ) {
   switch ( op1 & 0x0007 ) {
      case 0:
         flagT( !flagFE(  ) );
         break;
      case 1:
         flagT( !flagIE(  ) );
         break;
      case 2:
         flagT( !flagTE(  ) );
         break;
      case 3:
         flagT( !flagV(  ) );
         break;
      case 4:
         flagT( !flagZ0(  ) );
         break;
      case 5:
         flagT( !flagZ1(  ) );
         break;
      case 6:
         flagT( !flagPM(  ) );
         break;
      default:
         assert( false );
   }
   return SUCCESS;
}
CPUCore::errcode CPUCore::opXOR1( u16 op1 ) {
   int n = ( op1 & 0x0f00 ) >> 8;
   int i = op1 & 0x000f;
   u16 val = r( n ) ^ r( i );
   flagT( val == 0 );
   r( n, val );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opXOR2( u16 op1 ) {
   u16 imm = op1 & ( u16 ) 0x00ff;
   u16 val = r( 0 ) ^ imm;
   flagT( val == 0 );
   r( 0, val );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opXOR3( u16 op1, u16 op2 ) {
   int n = op1 & 0x000f;
   u16 val = r( n ) ^ op2;
   flagT( val == 0 );
   r( n, val );
   return SUCCESS;
}
CPUCore::errcode CPUCore::opEBKESEC( u16 op1 ) {
  u16 imm4 = op1 & 0x000f;
  int shift = (((op1>>4)-1) & 0x03) * 4;
  int mask = ~(0xF << shift);
  MSR[2] = (MSR[2] & mask) | (imm4 << shift);
  // FIXME: update behaviour
  return SUCCESS;
}
CPUCore::errcode CPUCore::opECLD( u16 op1 ) {
  int d = op1 & 0x7;
  int imm5 = ( op1 & 0x01f0 ) >> 4;
  if (imm5 >= 12)
    return RESERVED;
  int kind = imm5 / 4;
  int i = imm5 % 4;
  u16 val;
  if (op1 & 0x8) {
    val = r(d);
    switch (kind) {
    case 0:
      A[i] = (A[i] & 0xFFFF) | ((val & 0xFF) << 16);
      break;
    case 1:
      A[i] = (A[i] & 0xFF0000) | val;
      break;
    case 2:
      S[i] = val;
      break;
    }
  } else {
    switch (kind) {
    case 0:
      val = sext8to16(((u32) A[i]) >> 16);
      break;
    case 1:
      val = A[i];
      break;
    case 2:
      val = S[i];
      break;
    default:
      return RESERVED;
    }
    r(d, val);
  }
  return SUCCESS;
}
CPUCore::errcode CPUCore::opECR( u16 op1 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opEDIVQ( u16 op1 ) {
   return NOTIMPLEMENTED;
}

s32 CPUCore::subSD(int rpi, int si, bool dest) {
  assert (!flagMAC(XSD));
  return sext4to32(SD[rpi] >> (si*4 + (dest?8:0)) & 0xF);
}

u16 CPUCore::subRpis(int rps) {
  //FIXME: MC & modulo
  int rpi = rps & 0x3;
  int si = rps >> 2;
  u16 addr = RP[rpi];
  ADDRESS=RP[rpi]=addr+subSD(rpi, si, false);
  return addr;
}
u16 CPUCore::subRpid(int rpd) {
  //FIXME: MC & modulo
  int rpi = rpd & 0x3;
  int si = rpd >> 2;
  u16 addr = RP[rpi];
  ADDRESS=RP[rpi]=addr+subSD(rpi, si, true);
  return addr;
}
u16 CPUCore::subRpd(int i, unsigned ofs) {
  return (RPD[i] & (0xFFE0)) | ofs;
}

CPUCore::errcode CPUCore::subDspFlags(u32 val, bool add) {
  bool carry = add ? ((s32)val > 0x00FFFFFF) : ((s32)val >= 0);
  //  console.output("subDspFlags val=%x (%d); add=%d; carry=%d\n", val, val, add, carry);
  flagMAC(MAC_C, carry);
  return subDspFlagsLd(val);
}

CPUCore::errcode CPUCore::subDspFlagsLd(u32 val) {
  u16 msr0 = MSR[0] & 
    ~(
     (1<<MAC_Z) |
     (1<<MAC_V) |
     (1<<MAC_N)
    );
  msr0 |= (val == 0) ? (1<<MAC_Z) : 0;
  //FIXME: overflow flag
  msr0 |= (val & 0x800000) ? (1<<MAC_N) : 0;
  MSR[0] = msr0;
  return SUCCESS;
}


CPUCore::errcode CPUCore::opELD1a( u16 op1 ) {
  int t1 = ( ( op1 & 0x0060 ) >> 5 );
  int t2 = ( op1 & 0x0007 );
  A[t1] = gmini->memory.xread24(subRpis(t2));   //FIXME: x or y mem ?
  //  subDspFlags(A[t1]); flags not set when reading from mem
  return SUCCESS;
}

CPUCore::errcode CPUCore::opELD1b( u16 op1 ) {
   int t1 = ( ( op1 & 0x0060 ) >> 5 );
   s32 imm = sext5to32( ( op1 & 0x001f ) ) & BITS24;
   int op = ( op1 & 0x0180 ) >> 7;
   int tmp;
   switch (op) {
   case 0:
     subDspFlags(A[t1]-imm, false);
     return SUCCESS;
   case 2:
     A[t1] = imm;
     subDspFlagsLd(A[t1]);
     return SUCCESS;
   case 3:
     tmp = A[t1] + imm;
     subDspFlags(tmp, true);     
     A[t1] = tmp & BITS24;
     return SUCCESS;
   default:
     assert(false);
     return OOPS;     
   }
}



CPUCore::errcode CPUCore::opELD5a( u16 op1, u16 op2 ) {
  int mg = ( op1 & 0x00f8 ) >> 3;
  u24 imm = sext16to24(((op1 & 0x7) << 13) | (op2&0x1FFF));
  return subMg(mg, imm);
}

CPUCore::errcode CPUCore::subMg(int mg, u24 value) {
  u64 simm = value;
  int mi = (mg >> 1) & 0x01;
  switch(mg) {
  case 0x00:
  case 0x02:
    //    MA[0] = (MA[0] & 0x000000FFFFFFLL) | (simm << 24);
    MA[mi] = simm << 24;
    MAE(mi, simm >> 24);
    return SUCCESS;    
  case 0x01:
  case 0x03:
    MA[mi] = (MA[mi] & 0xFFFFFF000000LL) | value;
    return SUCCESS;    
  case 0x8:
  case 0x9:
  case 0xA:
  case 0xB:
    RP[mg-0x08] = value;
    return SUCCESS;    
  case 0xC:
  case 0xD:
    RPD[mg-0x0C] = value;
    return SUCCESS;    
  case 0xE:
  case 0xF:
    MC[mg-0x0E] = value;
    return SUCCESS;    
  case 0x10:
  case 0x11:
  case 0x12:
  case 0x13:
    SD[mg-0x10] = value;
    return SUCCESS;    
  case 0x14:  // "SA", "SI", "SG", "SR",
  case 0x15:
  case 0x16:
  case 0x17:
    S[mg-0x14] = value;
    return SUCCESS;
  case 0x18: //PH
    P = simm << 24;
    return SUCCESS;
  case 0x19: //PL
    P = (P & 0xFFFFFF000000LL) | value;
    return SUCCESS;
  case 0x1C:
  case 0x1D:
  case 0x1E:
    // MSRi
    MSR[mg-0x1C] = value;
    // FIXME: update behaviour according to flags
    return SUCCESS;    
  default:
    console.output(" subMG: %x\n", mg );
    exit(-1);
  }
}

CPUCore::errcode CPUCore::subMg1(int mg1, u24 value) {
  switch (mg1) {
  case 0x00:
  case 0x01:
  case 0x02:
  case 0x03:
    X[mg1] = value;
    return SUCCESS;
  case 0x04:
    MA[0] = (MA[0] & 0x000000FFFFFFLL) | ((u48)value << 24);
    return SUCCESS;    
  case 0x05:
    MA[0] = (MA[0] & 0xFFFFFF000000LL) | (u48)value;
    return SUCCESS;    
  case 0x06:
    MA[1] = (MA[1] & 0x000000FFFFFFLL) | ((u48)value << 24);
    return SUCCESS;    
  case 0x07:
    MA[1] = (MA[1] & 0xFFFFFF000000LL) | (u48)value;
    return SUCCESS;
  default:
    assert(false);
  }
  //fixme: flags
}

int CPUCore::subMg( int mg ) {
  switch(mg) {
  case 0x00: // MA0H
    return (MA[0] >> 24) & BITS24;
  case 0x01: // MA0L
    return MA[0] & BITS24;
  case 0x02: // MA1H
    return (MA[1] >> 24) & BITS24;
  case 0x03: // MA1L
    return (MA[1] & BITS24);

    //   "MA0SR", "MA0SL", "MA1SR", "MA1SL",

  case 0x8: // "RP0", "RP1", "RP2", "RP3",
  case 0x9:
  case 0xA:
  case 0xB:
    return RP[mg-0x08];
  case 0xC: // "RPD0", "RPD1",
  case 0xD:
    return RPD[mg-0x0C];
  case 0xE: //  "MC0", "MC1",
  case 0xF:
    return MC[mg-0x0E];
  case 0x10: // "SD0", "SD1", "SD2", "SD3",
  case 0x11:
  case 0x12:
  case 0x13:
    return SD[mg-0x10];
  case 0x14:  // "SA", "SI", "SG", "SR",
  case 0x15:
  case 0x16:
  case 0x17:
    return S[mg-0x14];
  case 0x18: //PH
    return (P >> 24) & BITS24;
  case 0x19: //PL
    return P & BITS24;
  case 0x1A: //MAiRN
  case 0x1B:
    return ((MA[mg-0x1A]+0x800000) >> 24) & BITS24;
  case 0x1C: //MSRi
  case 0x1D:
  case 0x1E:
    return MSR[mg-0x1C];
  case 0x1F: // PRN
    return ((P+0x800000) >> 24) & BITS24;
  default:
    console.output(" subMG: %x\n", mg );
    assert(false);
  }
}

u24 CPUCore::subPi(int pi) {
  if (pi) {
    return P & BITS24;
  } else {
    return (P >> 24) & BITS24;
  }
}

CPUCore::errcode CPUCore::opELD5b( u16 op1 ) {
  int t1 = ( op1 & 0x000c ) >> 2;
  int t2 = ( op1 & 0x0003 );
  S[t2] = S[t1];
  return SUCCESS;
}
CPUCore::errcode CPUCore::opELD5c( u16 op1 ) {
  int t1 = ( op1 & 0x0007 );
  int t2 = ( ( op1 & 0x0060 ) >> 5 );
  bool srg = (op1 & 0x0080);
  gmini->memory.xwrite24(subRpid(t1), srg?S[t2]:A[t2]);
  return SUCCESS;		       
}
CPUCore::errcode CPUCore::opELD5d( u16 op1 ) {
  int t1 = ( ( op1 & 0x0080 ) >> 7 );
  int adr5 = ( op1 & 0x001f );
  int t2 = ( ( op1 & 0x0060 ) >> 5 );
  bool write = ( op1 & 0x0100 );
  if (write)
    gmini->memory.xwrite24(subRpd(t1, adr5), A[t2]);
  else
    A[t2] = gmini->memory.xread24(subRpd(t1, adr5));
  //No flags updated when from memory
  return SUCCESS;
}
CPUCore::errcode CPUCore::opELD5e( u16 op1 ) {
  int imm5 = sext5to32(op1 & 0x001f);
  S[MAC_SA] = imm5;
  return SUCCESS;
}


CPUCore::errcode CPUCore::opELD5f( u16 op1, u16 op2 ) {
  int mgx = ( ( op1 & 0x0018 ) >> 3 );
  int imm16 = sext16to32( ((op1 & 0x7) << 13) | (op2&0x1FFF) );
  X[mgx] = imm16 & BITS24;
  return SUCCESS;
}

CPUCore::errcode CPUCore::opELD5g( u16 op1 ) {
  int mgx = ( op1 & 0x0003 );
  int an =  ( ( op1 & 0x000c ) >> 2 );
  X[mgx] = A[an];
  return SUCCESS;
}

CPUCore::errcode CPUCore::opELD5h( u16 op1 ) {
  int mg1 = ( op1 & 0x0007 );
  int pi = ( op1 & 0x0008 ) >> 3;
  if ( op1 & 0x0070 ) 
    return subMg1(mg1, subPi(pi));
  else
    assert (false); // opcode 111: subPi(pi, subMg1(mg1));
}

CPUCore::errcode CPUCore::opELD5i( u16 op1 ) {
  int mg = op1 & 0x001f;
  int an = ( op1 & 0x0060 ) >> 5;
  return subMg(mg, A[an]);
}

CPUCore::errcode CPUCore::opELD5j( u16 op1 ) {
  int mg1 =  ( ( op1 & 0x00e0 ) >> 5 );
  int rps = ( op1 & 0x0007 );
  
  return subMg1(mg1, gmini->memory.xread24(subRpis(rps)));
}


CPUCore::errcode CPUCore::opELD1( u16 op1, u16 op2 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opELD5( u16 op1 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opELD5( u16 op1, u16 op2 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opEMAD2( u16 op1 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opEMAD3( u16 op1 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opEMINMAX2( u16 op1 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::subEMINMAX( int op, u24& dst, u24 src ) {
  //OP == 0 : MAX
  int s = sext24to32(src), d = sext24to32(dst);
  bool assign;
  if (op == 0) {
    // MAX
    assign = s >= d;
    subDspFlags(dst-src, false);
    
  } else {
    // MIN
    assign = d >= s;
    subDspFlags(src-dst, false);
  }
  if (assign) {
    RP[3] = ADDRESS;
    dst = src;
  }
  return SUCCESS;
}

CPUCore::errcode CPUCore::opEMINMAX0( u16 op1 ) {
  int op = op1 & 1;
  int ai = (op1 >> 3) & 1;
  return subEMINMAX(op, A[ai], A[1-ai]);
}
CPUCore::errcode CPUCore::opEMINMAX1( u16 op1 ) {
  int op = op1 & 1;
  int ci = (op1 >> 2) & 1;
  int ai = (op1 >> 3) & 1;
  return subEMINMAX(op, A[ai], A[2+ci]);
}  
CPUCore::errcode CPUCore::opEMLD2( u16 op1 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opEMSB2( u16 op1 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::subEMUL( int xiyi ) {
  int xi = (xiyi >> 1) + 2;
  int yi = xiyi & 1;
  s64 x, y;
  bool usm = flagMAC1(USM);
  if (usm && (xi==1))
    x = X[xi];
  else
    x = sext24to32(X[xi]);
  
  if (usm && (yi==1))
    y = X[yi];
  else
    y = sext24to32(X[yi]);

  s64 p = x*y;
  //FIXME : overflow

  if (flagMAC1(PSH1)) {
    assert(p!=1); // otherwise overflow
    p = p >> 1;
  }
  
  P = p & BITS48;
  

  return SUCCESS;
}

CPUCore::errcode CPUCore::opEMUL1( u16 op1 ) {
  int xiyi  = ( op1 & 0x0003 );
  int op = ((op1 & 0x0008) >> 3) + ((op1 & 0x0080) >> 6);
  int mi = ( ( op1 & 0x0004 ) >> 2 );
  subMiP(mi,op);
  return subEMUL(xiyi);
  // no extra load/store 
}

CPUCore::errcode CPUCore::opEMUL0(u16 op1) {
  int mi = ( ( op1 & 0x0008 ) >> 3 );
  int xiyi = ( ( op1 & 0x0180 ) >> 7 );
  int mgx= ( ( op1 & 0x0060 ) >> 5 );
  int rps = ( op1 & 0x0007 );

  int op = ((op1&0x0200) >> 8) + ((op1&0x0010) >> 4);

  subMiP(mi,op);
  subEMUL(xiyi);

  X[mgx]=gmini->memory.xread24(subRpis(rps));
  return SUCCESS;
}

void CPUCore::subMiP(int mi, int op) {
  u64 mai, p;
  mai = ((u64)sext4to32(MAE(mi)) << 48) | MA[mi];
  p = sext48to64(P);
  switch (op) {
  case 0: //ADD
    mai += p;
    break; 
  case 1: //SUB
    mai -= p;
    break;
  case 2: //LD
    mai = p;
    break;
  case 3: //NOP
    break;
  default:
    assert(false);
  }
  MA[mi] = mai & BITS48;
  MAE(mi, mai >> 48); 
  //FIXME: extra overflow
}

void CPUCore::subAnOp(int an, int op, u24 val) {
  switch (op) {
  case 0: //ADD
    A[an]=(A[an]+val) & BITS24;
    break; 
  case 1: //SUB
    A[an]=(A[an]-val) & BITS24;
    break;
  case 2: //LD
    A[an]=val;
    break;
  default:
    assert(false);
  }
  //FIXME: flags
}


CPUCore::errcode CPUCore::opEMUL2( u16 op1 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opEMUL3( u16 op1 ) {
  int mi = ( op1 & 0x0008 ) >> 3 ;
  int xiyi = ( op1 & 0x0180 ) >> 7;
  int op = ((op1 & 0x0010) >> 4) + ((op1 & 0x0200) >> 8);
  subMiP(mi, op);
  subEMUL(xiyi);
  return opELD6(op1);
}

CPUCore::errcode CPUCore::opELD6( u16 op1 ) {
  int rp01 = op1 & 0x0001;
  int rp01i = (op1 & 0x0002) >> 1;
  int rp3i = ( op1 & 0x0004 ) >> 2;
  int xi = ( op1 & 0x0020 ) >> 5;
  int yi = ( op1 & 0x0040 ) >> 6;

  X[yi+0] = gmini->memory.xread24(subRpis((rp3i << 2) + 3));
  X[xi+2] = gmini->memory.xread24(subRpis((rp01i << 2) + rp01));
  
  return SUCCESS;
}

CPUCore::errcode CPUCore::opENMSK( u16 op1 ) {
  int imm4 = op1 & 0x000f;
  u24 mask = (((u24)BITS16) >> (16-imm4));
  S[MAC_SG] &= mask;
  //FIXME: flags
  return SUCCESS;
}
CPUCore::errcode CPUCore::opENOP( u16 op1 ) {
   return SUCCESS;
}


CPUCore::errcode CPUCore::opES( u16 op1 ) {
   int bs = op1 & 0x000F;
   bool set = op1 & 0x0010;
   switch (bs) {
   case 0x00:
   case 0x01:
   case 0x02:
   case 0x03:
     SETBIT(set, MSR[2], BK0+bs);
     return SUCCESS;     
   case 0x04:
   case 0x05:
   case 0x06:
   case 0x07:
     SETBIT(set, MSR[0], ME0+bs-0x04);
     return SUCCESS;          
   case 0x08:
   case 0x09:
   case 0x0A:
   case 0x0B:
   case 0x0C:
     return NOTIMPLEMENTED;
   case 0x0D:
     SETBIT(set, MSR[0], XSD);
     return SUCCESS;
   case 0x0E:
     SETBIT(set, MSR[1], PSH1);
     return SUCCESS;     
   case 0x0F:
     SETBIT(set, MSR[1], NQ);
     return SUCCESS;
   default:
     assert(false);
   }
     //FIXME: it does more than just setting a bit.
};


CPUCore::errcode CPUCore::opERPD( u16 op1 ) {
  int rpd = op1 & 0x0007;
  subRpid(rpd);
  return SUCCESS;
}
CPUCore::errcode CPUCore::opERPN1( u16 op1 ) {
  
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opERPN2( u16 op1, u16 op2 ) {
  int rpi = ( op1 & 0x0018 ) >> 3;
  int imm16 = ( ((op1 & 0x7) << 13) | (op2&0x1FFF) );

  RP[rpi] += imm16;
  ADDRESS = RP[rpi];
  
  //FIXME: modulo

  return SUCCESS;
}
CPUCore::errcode CPUCore::opERPR( u16 op1 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opERPS( u16 op1 ) {
  int rps = op1 & 0x0007;
  subRpis(rps);
  return SUCCESS;
}
CPUCore::errcode CPUCore::opESD( u16 op1 ) {
  assert (!flagMAC(XSD));
  int ns =  ( op1 & 0x0030 ) >> 4;
  unsigned imm4 = op1 & 0x000f;
  int i = ( op1 & 0x00c0 ) >> 6;
  int shift = ns*4;
  SD[i] = (SD[i] & ~(0xF<<shift)) | (imm4 << shift);
  return SUCCESS;
}

CPUCore::errcode CPUCore::opESFT( u16 op1 ) {
  int op = (op1 >> 4) & 0x3;
  int t1 = ( op1 & 0x000c ) >> 2;
  int t2 = ( op1 & 0x0003 );

  u24 asr; // shifted value
  int asa; // shift amount
  bool carry;
  
  switch (t1) {
  case 0:
  case 1:
    asr = A[t1];
    break;
  case 2:
    asr = S[MAC_SI];
    break;
  case 3:
    asr = S[MAC_SR];
    break;
  default:
    return OOPS;
  }
  switch (t2) {
  case 0:
  case 1:
    asa = A[t2];
    break;
  case 2:
    asa = S[MAC_SA];
    break;
  default:
    return RESERVED;
  }
  asa &= 0x7F; // only so many significant bits,
  asa = sext7to32(asa); // signed
  bool left = asa > 0; // direction ?
  asa = abs(asa);
  asa = asa > 48 ? 48 : asa; // saturate shift amount
  u64 r = asr; // extend value to 64
  u24 sr, sg; // results of the shift.
  if (left) {
    r = r << asa;
    sr = r & BITS24;
    sg = (r >> 24) & BITS24;
    carry = r & (1LL << 24);
  } else {
    r = r << (64-24); // enables to use native arithmetic shift
    if (op == 1)
      r = ((s64) r) >> asa; // arithmetic!
    else
      r = r >> asa;  // logic
    carry = r & (1LL << (64-24+1));
    sr = r >> (64-24);
    sg = (r >> (64-24-24)) & BITS24;
  }
  switch (op) {
    case 0: //logical
    case 1: //arithmetic
      S[MAC_SR] = sr;
      S[MAC_SG] = sg;
      break;
    case 2: //linked
      S[MAC_SR] |= sr;
      S[MAC_SG] |= sg;
      break;
    case 3: //double prec
      S[MAC_SG] |= sr;
      break;
  }
  subDspFlagsLd(sr);
  flagMAC(MAC_C, carry);
  // FIXME: other flags?
  return SUCCESS;
}
CPUCore::errcode CPUCore::opETST( u16 op1 ) {
   return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opReserved( u16 op1 ) {
   return RESERVED;
}
CPUCore::errcode CPUCore::opInvalid( u16 op1 ) {
   return INVALID;
}

CPUCore::errcode CPUCore::subEMOD0(int mod0, int n, int val ) {
  s32 tmp;
  s32 an = (A[n]);
  val = (val);
  switch(mod0) {
  case 0: // LD
    tmp = val;
    an = tmp;
    subDspFlagsLd(tmp);
    break;
  case 1: // ADD
    tmp = an + val;
    an = tmp;
    subDspFlags(tmp, true);
    break;
  case 2: //SUB
    tmp = an - val;
    an = tmp;
    subDspFlags(tmp, false);
    break;
  case 3: //CP
    tmp = an - val;
    subDspFlags(tmp, false);
    break;
  default:
    assert(false);
  }
  A[n] = an & BITS24;
  return SUCCESS;
}



CPUCore::errcode CPUCore::opEMOD0a( u16 op1, u16 op2 ) {
  int op = ( u8 ) ( ( op1 & 0x0060 ) >> 5 );
  int n = ( u8 ) ( ( op1 & 0x0018 ) >> 3 );
  u32 imm =  ((op1 & 0x7) << 13) | (op2&0x1FFF) ;
  return subEMOD0(op, n, sext16to24(imm));
}

CPUCore::errcode CPUCore::opEMOD0b(u16 op1) {
  int t1 = ( ( op1 & 0x0180 ) >> 7 );
  int t2 = ( ( op1 & 0x0060 ) >> 5 );
  int t3 = ( op1 & 0x001f );
  return subEMOD0(t1, t2, subMg(t3));
}

CPUCore::errcode CPUCore::opEMOD0c(u16 op1) {
  int mod0 =  ( ( op1 & 0x000c ) >> 2 );
  int an =  ( ( op1 & 0x0060 ) >> 5 );
  int t3 =  ( op1 & 0x0003 );
  bool mgx = op1 & 0x0010;
  if (mgx)
    return subEMOD0(mod0, an, X[t3]);
  else
    return subEMOD0(mod0, an, A[t3]);
}

CPUCore::errcode CPUCore::opEADDSUBLD1(u16 op1) {
  //11 EADD Mi,P, mgx,@rps 
  //12 ESUB Mi,P, mgx,@rps 
  //16 EADD Mi,P, An,@rps
  
  int mi = ( ( op1 & 0x0080 ) >> 7 );
  int mgx = ( ( op1 & 0x0060 ) >> 5 );
  int rps = ( op1 & 0x0007 );
  int op = ((op1 & 0x100) >> 8) + ((op1&0x800) >> 10);
  subMiP(mi, op);
  X[mgx] = gmini->memory.xread24(subRpis(rps));
  return SUCCESS;
}

CPUCore::errcode CPUCore::opEADDSUBLD2(u16 op1) {
  //13 	EADD Mi,P, An,@rps
  //14 	ESUB Mi,P, An,@rps
  //15 	ELD Mi,P, An,@rps 
  int mi = ( ( op1 & 0x0080 ) >> 7 );
  int an = ( ( op1 & 0x0060 ) >> 5 );
  int rps = ( u8 ) ( op1 & 0x0007 );
  int op = (op1 & 0x18) >> 3;
  subMiP(mi, op);
  A[an] = gmini->memory.xread24(subRpis(rps));
  return SUCCESS;
}

u24 CPUCore::subMga(int mga) {
  if (mga >= 2)
    return A[mga-2];
  else
    return MA[mga] >> 24;
}

CPUCore::errcode CPUCore::opEADDSUBLD3(u16 op1) {
// 17  	EADD Mi,P, @rpd,mga 
// 18 	ESUB Mi,P, @rpd,mga 
// 19 	ELD Mi,P, @rpd,mga

  int mi = ( ( op1 & 0x0080 ) >> 7 );
  int mga = ( ( op1 & 0x0060 ) >> 5 );
  int rpd = ( u8 ) ( op1 & 0x0007 );
  int op = (op1 & 0x18) >> 3;
  subMiP(mi, op);
  gmini->memory.xwrite24(subRpid(rpd), subMga(mga));
  return SUCCESS;
}

CPUCore::errcode CPUCore::opEADDSUBLD4(u16 op1) {
// 20  	EADD Mi,P, @rpd,P
// 21 	ESUB Mi,P, @rpd,P
// 22 	ELD Mi,P, @rpd,P 

  int mi = ( ( op1 & 0x0080 ) >> 7 );
  int rpd = ( u8 ) ( op1 & 0x0007 );
  int op = (op1 & 0x18) >> 3;
  subMiP(mi, op);
  gmini->memory.xwrite24(subRpid(rpd), P);
  return SUCCESS;
}

CPUCore::errcode CPUCore::opEADDSUBLD5(u16 op1) {
  return NOTIMPLEMENTED;
}

CPUCore::errcode CPUCore::opEADDSUBLD6(u16 op1) {
  return NOTIMPLEMENTED;
}

CPUCore::errcode CPUCore::opEADDSUBLD7(u16 op1) {
  return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opEADDSUBLD8(u16 op1) {
  return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opEADDSUBLD9(u16 op1) {
  //40 Ai,Ci, Cj,@rps
// 41 ESUB Ai,Ci, Cj,@rps
// 42 ELD Ai,Ci, Cj,@rps
  int ai =  ( ( op1 & 0x0008 ) >> 3 );
  int ci1 =  ( ( op1 & 0x0020 ) >> 5 );
  int ci2 = ( ( op1 & 0x0010 ) >> 4 );
  int rps = ( op1 & 0x0007 );
  int op = (op1 & 0xC0) >> 6;
  subAnOp(ai, op, A[2+ci1]);
  A[2+ci2] = gmini->memory.xread24(subRpis(rps));
  return SUCCESS;
}

CPUCore::errcode CPUCore::opEMOD1(u16 op1) {
  int n = ( ( op1 & 0x0060 ) >> 5 );
  bool t = op1 & 0x0010;
  int mod1 = ( op1 & 0x000f );
  u32 val;
  if (t && !flagMAC(MAC_T))
    return SUCCESS;

//   "ESRA", "ESLA", "ESRA8", "ESLA8",
//     "ESRC", "ESLC", "EINCC", "EDECC",
//     "ENEG", "EABS", "EFS16", "EFZ16",
//     "EFS8", "EFZ8", "EEXP", "EEXPC"
  switch(mod1) {
  case 0x0: //ESRA
    A[n] = (sext24to32(A[n]) >> 1) & BITS24;
    return SUCCESS;    
  case 0x1: // ESLA1
    A[n] = (A[n] << 1) & BITS24;
    return SUCCESS;
  case 0x2: // ESRA8
    A[n] = (sext24to32(A[n]) >> 8) & BITS24;
    return SUCCESS;    
  case 0x3: // ESLA8
    A[n] = (A[n] << 8) & BITS24;
    return SUCCESS;
  case 0x8: //ENEG
    A[n] = (-A[n]) & BITS24;
    return SUCCESS;    
  case 0x9: //EABS
    A[n] = abs(sext24to32(A[n])) & BITS24;
    return SUCCESS;    
  case 0xA: //EFS16
    A[n] = sext8to32(A[n]) & BITS24;
    subDspFlagsLd(A[n]);
    flagMAC(MAC_C, false);
    flagMAC(MAC_V, false);
    return SUCCESS;    
  case 0xB: //EFZ16
    A[n] = A[n] & BITS8;
    subDspFlagsLd(A[n]);
    flagMAC(MAC_C, false);
    flagMAC(MAC_V, false);
    return SUCCESS;    
  case 0xC: //EFS8
    A[n] = sext16to32(A[n]) & BITS24;
    subDspFlagsLd(A[n]);
    flagMAC(MAC_C, false);
    flagMAC(MAC_V, false);
    return SUCCESS;    
  case 0xD: //EFZ8
    A[n] = A[n] & BITS16;
    subDspFlagsLd(A[n]);
    flagMAC(MAC_C, false);
    flagMAC(MAC_V, false);
    return SUCCESS;    
  case 0xE: //EEXP
  case 0xF: //EEXPC
    {
      val = A[n];
      bool first = mod1 == 0xE;
      bool sign;
      bool carry = val & 1;
      if (first)
	sign = val & (1 << 23);
      else
	sign = flagMAC(MAC_C);
      int exp = 0;
      if (first || flagMAC(MAC_VS))
	while (exp < 24 && (((val & (1 << 23)) != 0)==sign)) {
	  val <<= 1;
	  exp++;
	}
      if (first) {
	S[MAC_SA] = exp-1;
      } else {
	S[MAC_SA] += exp;
      }
      flagMAC(MAC_C, carry);
      flagMAC(MAC_VS, exp == 24);
      flagMAC(MAC_N, false);
      flagMAC(MAC_Z, exp == 0);
      
      return SUCCESS;
    }
  default:
    return NOTIMPLEMENTED;
  }
  //FIXME: flags
}

CPUCore::errcode CPUCore::opEMOD2(u16 op1) {
  int mi =  ( ( op1 & 0x0010 ) >> 4 );
  int op =  ( op1 & 0x000f );
  switch (op) {
  case 0x0:     //  "ESRA  MA0"
    MA[mi] = (sext48to64(MA[mi]) >> 1) & BITS48;
    return SUCCESS;
  case 0x1:     //  "ESLA  MA0"
    MA[mi] = (MA[mi] << 1) & BITS48;
    return SUCCESS;    
//  "ERND  MA0"
//  "ECR   MA0"
//  "ESAT  MA0"
//  "ERESR MA0"
//  "Reserved"
//  "Reserved"
  case 0x08:
//  "ELD   MA0 MA1'"
    MA[mi] = MA[1-mi];
    MAE(mi, MAE(1-mi));
    return SUCCESS;
//  "Reserved"
  case 0x0A:
//  "EADD  MA0 P"
    subMiP(mi, 0); // 0: ADD
    return SUCCESS;
  case 0x0B:
    subMiP(mi, 1); // 0: SUB
    return SUCCESS;
//  "EADD  MA0 PSH"
//  "ESUB  MA0 PSH"
//  "EDIVQ MA0 P"
//  "Reserved"    
  default:
    return NOTIMPLEMENTED;
  }
}


