/*  Project: Open Source Gmini Tools
    file: CPUCore.h, header file for CPUCore.cpp

    Copyright (C) 2004  The gmEmu Project <gmemu@donat.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef CPUCoreH
#define CPUCoreH
#include <map>
#include <vector>
#include "Reference.h"

struct StackFrame {
   u22 called;
   u22 from;
   bool irq;
   int irqIndex;
      StackFrame(  ):irq( false ) {
}};

#define SETBIT(set, val, idx)			\
{						\
  if (set) {					\
    (val) |= (1<<(idx));			\
  } else {					\
    (val) &= ~(1<<(idx));			\
  }						\
}


class CPUCore {
 private:
   enum SPSRegister { SPCL_FIQ = 0, SPCH_FIQ = 1, SSR_FIQ = 2,
      SPCL_IRQ = 4, SPCH_IRQ = 5, SSR_IRQ = 6,
      SSR_SWI = 10, SSR_END
   };
   enum errcode { SUCCESS, ABSJUMP, RELJUMPUP, RELJUMPDN, SUCCESS_END, 
		  NOTIMPLEMENTED, INVALID, RESERVED,
		  OPSIZE, MEMORY, 
		  READONLY, WRITEONLY, KNOWNDATA, OOPS, BREAKPOINT, ERRCODE_END
   };
   enum MSR0_BITS {MAC_T, MAC_C, MAC_Z, MAC_N, MAC_V=5,
		   MAC_VS,
		   OP=9,XSD=10, ME0=12, ME1, ME2, ME3};
   enum MSR1_BITS {VM0, VM1, MV=2,OPM=3,USM=4,PSH1=5,NQ=6,OPMA=7};
   enum MSR2_BITS {BK0=12, BK1, BK2, BK3};
   static const char errcodestr[ERRCODE_END][20];
   static const char stateREGfile[14];

//   enum { OPHISTORY = 100 };
   enum { MAC_SA, MAC_SI, MAC_SG, MAC_SR, MAC_S_END };
// enum { DISABLED, ENABLED } state_t;
   enum reason_t { BREAK, REACHED, ERROR, SIGNAL };
   Emulator *gmini;
   u16 R[16];
   u6 E[8];
   u22 PCC;
   u16 SR, SPSR[SSR_END];

   //CalmMAC24 registers
   // 24 bit
   u24 X[4]; // Y0, Y1, X0, X1
   u24 S[4];
   u24 A[4]; // A, B, C, D

   // 16 bit
   u16 ADDRESS; //invisible register
   u16 MSR[3];
   u16 RP[4];
   u16 SD[MAC_S_END];
   u16 SD0E, SD3E;
   u16 RPD[2];
   u16 MC[2];

   u48 P;                       // 48 bit (PH PL)
   u52 MA[2];               // 52 bit (MSR1_MAi MAiH MAiL )
   bool EC3; 


   bool flagMAC(MSR0_BITS flag) {
     return MSR[0] & (1<<flag);
   }

   void flagMAC(MSR0_BITS flag, bool val) {
     if (val)
       MSR[0] |= (1<<flag);
     else
       MSR[0] &= ~(1<<flag);
   }

   bool flagMAC1(MSR1_BITS flag) {
     return MSR[1] & (1<<flag);
   }

   void flagMAC1(MSR1_BITS flag, bool val) {
     if (val)
       MSR[1] |= (1<<flag);
     else
       MSR[1] &= ~(1<<flag);
   }

   int MAE(int i) {
     return (MSR[1] >> (8 + 4*i)) & 0xF;
   }

   void MAE(int i, int val) {
     int sh = (8 + 4*i);
     MSR[1] = (MSR[1] & ((~0xF) << sh)) | ((val & 0xF) << sh);
   }

   bool flagT(  ) {
      return SR & 0x8000;
   };
   void flagT( bool t ) {
      if( t )
         SR |= 0x8000;
      else
         SR &= 0x7fff;
   }
   bool flagPM(  ) {
      return SR & 0x0040;
   }
   void flagPM( bool t ) {
      if( t )
         SR |= 0x0040;
      else
         SR &= 0xffbf;
   }
   bool flagZ1(  ) {
      return SR & 0x0020;
   }
   void flagZ1( bool t ) {
      if( t )
         SR |= 0x0020;
      else
         SR &= 0xffdf;
   }
   bool flagZ0(  ) {
      return SR & 0x0010;
   }
   void flagZ0( bool t ) {
      if( t )
         SR |= 0x0010;
      else
         SR &= 0xffef;
   }
   bool flagV(  ) {
      return SR & 0x0008;
   }
   void flagV( bool t ) {
      if( t )
         SR |= 0x0008;
      else
         SR &= 0xfff7;
   }
   bool flagTE(  ) {
      return SR & 0x0004;
   }
   void flagTE( bool t ) {
      if( t )
         SR |= 0x0004;
      else
         SR &= 0xfffb;
   }
   bool flagIE(  ) {
      return SR & 0x0002;
   }
   void flagIE( bool t ) {
      if( t )
         SR |= 0x0002;
      else
         SR &= 0xfffd;
   }
   bool flagFE(  ) {
      return SR & 0x0001;
   }
   void flagFE( bool t ) {
      if( t )
         SR |= 0x0001;
      else
         SR &= 0xfffe;
   }
   u16 r( int n ) {
      //assert( n >= 0 && n <= 15 );
      return R[n];
   }
   void r( int n, u16 val ) {
      //assert( n >= 0 && n <= 15 );
      R[n] = val;
   }
   u6 e( int n ) {
      //assert( n >= 8 && n <= 15 );
      return ( u6 ) ( E[n - 8] & 0x3f );
   }
   void e( int n, u6 val ) {
      //assert( n >= 8 && n <= 15 );
      E[n - 8] = ( u6 ) ( val & 0x3f );
   }
   u22 a( int n ) {
      //assert( n >= 8 && n <= 15 );
      return ( ( u22 ) e( n ) << 16 ) | ( u22 ) r( n );
   }
   void a( int n, u22 val ) {
      //assert( n >= 8 && n <= 15 );
      r( n, ( u16 ) ( val & 0x00ffff ) );
      e( n, ( u6 ) ( ( val & 0x3f0000 ) >> 16 ) );
   }
   void a( int n, u6 vale, u16 valr ) {
      //assert( n >= 8 && n <= 15 );
      e( n, vale );
      r( n, valr );
   }
   u16 d( int n ) {
      //assert( n >= 0 && n <= 7 );
      return r( n );
   }
   void d( int n, u16 val ) {
      //assert( n >= 0 && n <= 7 );
      r( n, val );
   }
 public:
   void setR( int n, u16 val );
   u32 getData( Location loc );

 private:
   u32 steps;
   bool stop;
   bool haveDebugBreak;
   u32 debugbreak;
   bool haveLCDBreak;
   u32 lcdbreak;
   u22 emudebugbreak;
   u22 fprfuncbreak;

   int printOp( u32 addr, char *opstring );
 public:
   enum Versions { gm120pre, gm120v120, gm120v130, gm120v140, gm120v150,
      gm120v170, gm120v180, gm120v190, gm120v193, gm120v1100,
      gm220v160, gm220v170, gm220v180, gm220v190,
      gmSPv110, gmSPv120, gmSPv123, gmSPv130
   };
   enum Overides { ovFlagT = 0, ovFlagPM, ovFlagZ1, ovFlagZ0, ovFlagV, ovFlagTE, ovFlagIE, ovFlagFE = 7,
      ovR0 = 8, ovR1, ovR2, ovR3, ovR4, ovR5, ovR6, ovR7 = 15,
      ovR8 = 16, ovR9, ovR10, ovR11, ovR12, ovR13, ovR14, ovR15 = 23,
      ovE8 = 24, ovE9, ovE10, ovE11, ovE12, ovE13, ovE14, ovE15 = 31,
      ovA8 = 32, ovA9, ovA10, ovA11, ovA12, ovA13, ovA14, ovA15 = 39
   };
   void overide( Overides what, u32 value );
 private:
   void doOveride(  );
   void timerIncrement( int index, int CON, int CNT, int DATA );
   void raiseIRQ( int index );
   Versions version;
 public:
   void setVersion( Versions ver ) {
      version = ver;
   }
   void disasm( u22 address );
   CPUCore( Emulator * gm );
   void locatefunctions(  );
   void restoreState(  );
   void saveState(  );
   void dumpregs(  );
   void dumpops( ostream & o );
   void dumpops(  );
   void dump(  );
   void dumpBacktrace( ostream & o );
   void go( u22 address = 0 );
   errcode step(  );
   void stepOver(  );
   void walk( u22 address = 0xffffffff );
   void jog( u22 address = 0xffffffff );
   void run( u22 address = 0xffffffff );
   void run2ret(  );
   void sigint(  );
   vector < StackFrame > callstack;
 private:
   void stopping( reason_t why, char *what );
   void doTrace(  );
   errcode delaySlot(  );
   errcode branch( s22 off, bool save = false, bool delay = false );
   errcode jump( u22 addr, bool save = false, bool delay = false );
   errcode subRet( u22 addr, bool delay );
   errcode doop( int mnnum, u16 op, u16 op2 = ZERO );
   void CPUCore::call( u22 addr );
   u16 sub2scomp( u16 val );

   inline errcode subADSBC( int n, u16 v2, bool c, bool m );
   inline errcode subSBC( int n, u16 v2, bool c );
   inline errcode subSUB( int n, u16 v2 );
   inline errcode subAND( int n, u16 v2 );
   inline errcode opADC1( u16 op1 );   // 19
   inline errcode opADC2( u16 op1, u16 op2 ); // 52
   inline errcode opADD1( u16 op1 );   // 17
   inline errcode opADD2( u16 op1 );   // 0
   inline errcode opADD3( u16 op1, u16 op2 ); // 49
   inline errcode opADD4( u16 op1 );   // 138
   inline errcode opADD5( u16 op1, u16 op2 ); // 50
   inline errcode opADD6( u16 op1 );   // 134
   inline errcode opAND1( u16 op1 );   // 21
   inline errcode opAND2( u16 op1 );   // 66
   inline errcode opAND3( u16 op1, u16 op2 ); // 54
   inline errcode opBIT( u16 op1 );    // 72-75
   inline errcode opBNZD( u16 op1 );   // 154
   inline errcode opBR( u16 op1 );     // 156-161
   inline errcode opBRAEC( u16 op1 );  // 151,153
   inline errcode opBREAK( u16 op1 );  // 82
   inline errcode opBSRD( u16 op1 );   // 150
   inline errcode opCLRSR( u16 op1 );  // 78
   inline errcode opCMP1( u16 op1 );   // 25-29
   inline errcode opCMP2( u16 op1, u16 op2 ); // 58-62
   inline errcode opCMP3( u16 op1 );   // 133
   inline errcode opCMPEQ1( u16 op1 ); // 125
   inline errcode opCMPEQ2( u16 op1 ); // 65
   inline errcode opCMPEQ3( u16 op1, u16 op2 );  // 136
   inline errcode opCOM( u16 op1 );    // 45-47
   inline errcode opDECC( u16 op1 );   // 44
   inline errcode opDT( u16 op1 );     // 42
   inline errcode opEXT( u16 op1 );    // 48
   inline errcode opINCC( u16 op1 );   // 43
   inline errcode opJMP1( u16 op1 );   // 37-40
   inline errcode opJMP2( u16 op1, u16 op2 ); // 117-120
   inline errcode opLD1( u16 op1 );    // 30
   inline errcode opLD2( u16 op1 );    // 126
   inline errcode opLD3( u16 op1 );    // 2
   inline errcode opLD4( u16 op1, u16 op2 );  // 63
   inline errcode opLD5( u16 op1, u16 op2 );  // 137
   inline errcode opLDRE( u16 op1 );   // 123-124
   inline errcode opLDB1( u16 op1 );   // 9,13
   inline errcode opLDB2( u16 op1, u16 op2 ); // 129-130
   inline errcode opLDB3( u16 op1 );   // 10,14
   inline errcode opLDB4( u16 op1 );   // 70-71
   inline errcode opLDC( u16 op1 );    // 121
   inline errcode opLDPC( u16 op1 );   // 115
   inline errcode opLDSvR1( u16 op1 ); // 89-91,93-95,98
   inline errcode opLDSvR2( u16 op1 ); // 101-103,105-107,110
   inline errcode opLDSR( u16 op1 );   // 83-84
   inline errcode opLDW1( u16 op1 );   // 3-4
   inline errcode opLDW2( u16 op1 );   // 5,7
   inline errcode opLDW3( u16 op1, u16 op2 ); // 127-128
   inline errcode opLDW4( u16 op1 );   // 6,8
   inline errcode opLDW5( u16 op1 );   // 11,15
   inline errcode opLDW6( u16 op1, u16 op2 ); // 131-132
   inline errcode opLDW7( u16 op1 );   // 12,16
   inline errcode opMUL( u16 op1 );    // 140-143
   inline errcode opNOP( u16 op1 );    // 81
   inline errcode opOR1( u16 op1 );    // 22
   inline errcode opOR2( u16 op1 );    // 67
   inline errcode opOR3( u16 op1, u16 op2 );  // 55
   inline errcode opPOP1( u16 op1 );   // 144
   inline errcode opPOP2( u16 op1 );   // 146
   inline errcode opPUSH1( u16 op1 );  // 147
   inline errcode opPUSH2( u16 op1 );  // 149
   inline errcode opRETD( u16 op1 );   // 88
   inline errcode opRET_FIQ( u16 op1 ); // 85
   inline errcode opRET_IRQ( u16 op1 ); // 86
   inline errcode opRET_SWI( u16 op1 ); // 87
   inline errcode opRL( u16 op1 );     // 32
   inline errcode opRR( u16 op1 );     // 31
   inline errcode opRRC( u16 op1 );    // 33
   inline errcode opSBC1( u16 op1 );   // 20
   inline errcode opSBC2( u16 op1, u16 op2 ); // 53
   inline errcode opSETSR( u16 op1 );  // 79
   inline errcode opSLB( u16 op1 );    // 41
   inline errcode opSR( u16 op1 );     // 35
   inline errcode opSRA( u16 op1 );    // 36
   inline errcode opSRB( u16 op1 );    // 34
   inline errcode opSUB1( u16 op1 );   // 18
   inline errcode opSUB2( u16 op1 );   // 1
   inline errcode opSUB3( u16 op1 );   // 139
   inline errcode opSUB4( u16 op1, u16 op2 ); // 51
   inline errcode opSUB5( u16 op1 );   // 135
   inline errcode opSWI( u16 op1 );    // 77
   inline errcode opSYS( u16 op1 );    // 76
   inline errcode opTST1( u16 op1 );   // 24
   inline errcode opTST2( u16 op1 );   // 69
   inline errcode opTST3( u16 op1, u16 op2 ); // 57
   inline errcode opTSTSR( u16 op1 );  // 80
   inline errcode opXOR1( u16 op1 );   // 23
   inline errcode opXOR2( u16 op1 );   // 68
   inline errcode opXOR3( u16 op1, u16 op2 ); // 56
   inline errcode opEBKESEC( u16 op1 );    // 279
   inline errcode opECLD( u16 op1 );   // 164
   inline errcode opECR( u16 op1 );    // 272.3
   inline errcode opEDIVQ( u16 op1 );  // 272.E
   inline errcode opELD1( u16 op1, u16 op2 ); // 166.0, 169 /////CHECK THIS OUT!!!!
   inline errcode opELD5( u16 op1, u16 op2 ); // 165, 167, 170 //CHECK THIS OUT!!!!
   inline errcode opELD5( u16 op1 );   // 194, 202, 209, 211, 212, 213, 214, 226, 228,
   // 229, 230, 236, 240, 250, 251, 275, 276, 277, 278
   inline errcode opELD1a( u16 op1);
   inline errcode opELD5a( u16 op1, u16 op2 );
   inline errcode opELD5b( u16 op1 );
   inline errcode opELD5c( u16 op1 );
   inline errcode opELD6( u16 op1 );   // 219
   inline errcode opEMAD2( u16 op1 );  // 171, 241, 252, 258
   inline errcode opEMAD3( u16 op1 );  // 215
   inline errcode opEMINMAX0( u16 op1 );
   inline errcode opEMINMAX1( u16 op1 );
   inline errcode opEMINMAX2( u16 op1 );
   inline errcode opEMLD2( u16 op1 );  // 173, 243, 254, 260
   inline errcode opEMSB2( u16 op1 );  // 172, 242, 253, 259
   inline errcode opEMUL0( u16 op1 ); 
   inline errcode opEMUL1( u16 op1 );  // 263
   inline errcode opEMUL2( u16 op1 );  // 174, 244, 255, 261
   inline errcode opEMUL3( u16 op1 );  // 218
   inline errcode opENMSK( u16 op1 );  // 227
   inline errcode opENOP( u16 op1 );   // 249
   inline errcode opERPD( u16 op1 );   // 271
   inline errcode opERPN2( u16 op1, u16 op2 ); // 168
   inline errcode opERPN1( u16 op1 );   // 269
   inline errcode opERPR( u16 op1 );   // 264
   inline errcode opERPS( u16 op1 );   // 270
   inline errcode opES( u16 op1 );     // 274
   inline errcode opESD( u16 op1 );    // 231, 232, 233, 234
   inline errcode opESFT( u16 op1 );   // 222
   inline errcode opETST( u16 op1 );   // 273
   inline errcode opReserved( u16 op1 );   // 64, 92, 96, 97, 99, 100, 104, 108, 109,
   // 111, 112, 113,
   // 114, 116, 122, 145, 148, 152, 155, 187, 195, 203, 220,
   // 221, 272.6, 272.7, 272.9, 272.F
   inline errcode opInvalid( u16 op1 );

   inline errcode opEMOD0a( u16 op1, u16 op2 );
   inline errcode opEMOD0b( u16 op1 );
   inline errcode opEMOD0c( u16 op1 );
   inline errcode opEMOD1( u16 op1 );
   inline errcode opEMOD2(u16 op1);
 
   inline errcode opELD1b( u16 op1 );
   inline errcode opELD5d( u16 op1 );
   inline errcode opELD5e( u16 op1 );
   inline errcode opELD5f( u16 op1, u16 op2 );
   inline errcode opELD5g( u16 op1 );
   inline errcode opELD5h( u16 op1 );
   inline errcode opELD5i( u16 op1 );
   inline errcode opELD5j( u16 op1 );

   inline errcode subEMUL( int xiyi );
   inline errcode subEMOD0(int mod0, int n, int val );
   inline errcode opEADDSUBLD1(u16 op1);
   inline errcode opEADDSUBLD2(u16 op1);
   inline errcode opEADDSUBLD3(u16 op1);
   inline errcode opEADDSUBLD4(u16 op1);
   inline errcode opEADDSUBLD5(u16 op1);
   inline errcode opEADDSUBLD6(u16 op1);
   inline errcode opEADDSUBLD7(u16 op1);
   inline errcode opEADDSUBLD8(u16 op1);
   inline errcode opEADDSUBLD9(u16 op1);

   inline errcode subEMINMAX(int, u24&, u24);
   
   void subMiP(int mi, int op);
   void subAnOp(int mi, int op, u24 val);
   inline errcode subDspFlags(u24 val, bool add);
   inline errcode subDspFlagsLd(u24 val);

   u16 subRpis(int rps);
   u16 subRpid(int rpd);
   u16 subRpd(int i, unsigned ofs);
   int subSD(int rpi, int si, bool dest);

   int subMg( int mg );
   inline errcode subMg(int mg, u24 value);

   inline errcode subMg1(int mg1, u24 value);

   u24 subMga( int mga );

   u24 subPi(int pi);
};
#endif
