
#include "Emulator.h"
#include "trace.h"
#include <assert.h>

I2CBus::I2CBus( Emulator * gm ):gmini( gm ) {
  masterState[SCL] = masterState[SDA] = true;
  slaveState[SCL] = slaveState[SDA] = true;
  lineState[SCL] = lineState[SDA] = true;
  transfer = false;
  byteIndex = 0;
  bitIndex = 0;
};

void I2CBus::lineInteractMaster( int line, bool write, bool & state ) {
   if( write ) {
      masterState[line] = state;
   }
   else {
      // line naturally goes up.
      masterState[line] = true;
   }

   // AND circuit on each line
   bool newState = masterState[line] && slaveState[line];

   if( newState != lineState[line] ) {
      lineState[line] = newState;
      if( TRACE_I2C )
         gmini->tracer.log( "I2C %s %d %s", ( line == SCL ) ? "SCL" : "SDA", newState, newState ? "/" : "\\" );
      lineNotifySlave( line, newState );
      if( TRACE_I2C ) {
	gmini->tracer.log("I2C slave : SCL=%d SDA=%d trans=%d byteindex=%d", slaveState[SCL], slaveState[SDA], transfer, byteIndex );
	gmini->tracer.log("I2C bus   : SCL=%d SDA=%d", lineState[SCL], lineState[SDA] );     
      }
   }

   // return linestate if requested
   if( !write )
      state = lineState[line];
}

void I2CBus::setSlaveState( int line, bool state ) {
   slaveState[line] = state;
   // lineState[line] = masterState[line] && slaveState[line];
}

void I2CBus::lineNotifySlave( int line, bool state ) {
   if( line == SCL ) {
      // SCL changes
      if( !transfer ) {
         // if no transfer going, release SDA & ignore:
         setSlaveState( SDA, true );
         return;
      }
      if( state ) {
         // clock goes up
         if( direction == 0 ) {
            // master => slave
            if( bitIndex == 8 ) {
               // slave send ACK
               if( TRACE_I2C ) {
                  gmini->tracer.log( "I2C slave ACK being read." );
                  gmini->tracer.log( "I2C BYTE to slave: %02x", byteBuffer );
               }
               if( byteIndex == 0 ) {
                  slaveAddress = byteBuffer & 0xFE;
                  direction = byteBuffer & 0x01;
                  if( TRACE_I2C )
                     gmini->tracer.log( "I2C ADDRESS: slave=%x, dir=%d", slaveAddress, direction );
               }
	       assert(byteIndex < I2C_MAXTRANSFER - 1);
               transferBuffer[byteIndex] = byteBuffer;
               byteIndex++;
               bitIndex = 0;
               if( direction == 1 ) {
                  // if changed direction, prepare slave data
                  requestSlaveData(  );
                  byteBuffer = transferBuffer[byteIndex];
               }
            }
            else {
               // slave read bit
               if( TRACE_I2C )
                  gmini->tracer.log( "I2C got bit %d=%d", bitIndex, lineState[SDA] );
               bitIndex++;
               byteBuffer = ( byteBuffer << 1 ) + ( lineState[SDA] ? 1 : 0 );
            }
         }
         else {
            // slave => master
            if( bitIndex == 0 ) {
               if( lineState[SDA] ) {
                  // NACK from master!
                  transfer = false;
                  if( TRACE_I2C ) {
                     char buf[2048];
                     char *p = buf;
                     *p = 0;
                     for( int i = 1; i < byteIndex; i++ ) {
                        p += strlen( p );
                        sprintf( p, "%02x", transferBuffer[i] );
                     }
                     gmini->tracer.log( "I2C NACK: slave=%x, dir=%d data=%s", slaveAddress, direction, buf );
                  }
               }
               else {
                  // ACK from master 
                  if( TRACE_I2C ) {
                     gmini->tracer.log( "I2C master ACK" );
                  }
               }
            }
         }
      }
      else {
         // clock goes down.
         if( direction == 0 ) {
            if( bitIndex == 8 ) {
               // send ACK
               setSlaveState( SDA, false );
            }
            else {
               // release whatever ack provided.
               setSlaveState( SDA, true );
            }
         }
         else {
            if( bitIndex == 8 ) {
               // release & wait for ack
               setSlaveState( SDA, true );
               bitIndex = 0;
               if( TRACE_I2C )
                  gmini->tracer.log( "I2C BYTE to master over: %02x, %02x", byteIndex, transferBuffer[byteIndex] );
               // load next byte
               byteIndex++;
	       assert(byteIndex < I2C_MAXTRANSFER);
               byteBuffer = transferBuffer[byteIndex];
            }
            else {
               // slave sends bit.
               bool bit = ( byteBuffer & 0x80 ) != 0;
               setSlaveState( SDA, bit );
               if( TRACE_I2C )
                  gmini->tracer.log( "I2C sent bit %d, %d", bitIndex, bit );
               // next bit.
               byteBuffer = byteBuffer << 1;
               bitIndex++;
            }
         }
      }
   }
   else {
      assert( line == SDA );
      if( lineState[SCL] ) {
         // data change when clock up: special condition
         if( state ) {
            // Master sends STOP
            if( TRACE_I2C )
               gmini->tracer.log( "I2C stop" );
	    if (transfer) {
	      // Report & process data as applicable
	      if( TRACE_I2C_TRANSFERS ) {
		char buf[I2C_MAXTRANSFER*2+1];
		char *p = buf;
		*p = 0;
		for( int i = 1; i < byteIndex; i++ ) {
                  p += sprintf( p, "%02x", transferBuffer[i] );
		}
		gmini->tracer.log( "I2C TRANSFER: slave=%x, dir=%d data=%s", slaveAddress, direction, buf );
	      }
	      if( direction == 0 )
		feedSlaveData(  );
	      transfer=false;
	    }
         }
         else {
            // Master sends START
            if( TRACE_I2C )
               gmini->tracer.log( "I2C start" );
	    slaveAddress = 0;
            transfer = true;
            bitIndex = 0;
            byteIndex = 0;
            direction = 0;
         }
      }
   }
}

u8 rtcClockData[0x14] = { 0x00, // csec
   0x33,                        // sec + stop bit
   0x22,                        // minutes
   0x11,                        // hours + century bits
   0x03,                        // day of week
   0x07,                        // day of month
   0x12,                        // month
   0x04,                        // year
   0x00,
   0x00,
   0x00,
   0x00,
   0x00,
   0x00,
   0x00,
   0x00,
   0x00,
   0x00,
   0x00,
   0x00
};

void I2CBus::requestSlaveData(  ) {
   assert( slaveAddress == 0xD0 );
   // RTC Clock
   for( int i = 0; i < 0x14; i++ ) {
      transferBuffer[i + 1] = rtcClockData[i];
   }
}


void I2CBus::processAudioCommand(u8* data) {
  u16 cmd = ((u16)data[1]) | (((u16)data[0]) << 8);
  int val = cmd & 0x1FF;
  int reg = cmd >> 9;
  if (TRACE_AUDIO)
    gmini->tracer.log("AUDIO command reg=%x, val=%x\n", reg, val);
}

void I2CBus::feedSlaveData(  ) {
   // skip 1st byte: slave address + direction
   u8 *data = transferBuffer + 1;
   int size = byteIndex - 1;

   switch ( slaveAddress ) {
   case 0:
     // dummy transfer (rockbox i2c init)
     break;
      case 0x2B:
         // SPIDF receiver
         break;
      case 0x36:
         // Audio codec
	assert(size == 2);
	processAudioCommand(data);
         break;
      case 0xD0:
         // RTC Clock
         assert( size <= 0x14 );
         for( int i = 0; i < size; i++ ) {
            rtcClockData[i] = data[i];
         }
         break;
   }
}
