/*  Project: Open Source Gmini Tools
    file: CommentMap.cpp, mantains a dynamic comment mapping

    Copyright (C) 2004  The gmEmu Project <gmemu@donat.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <fstream>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "CommentMap.h"
#include "SymMap.h"
#include "Console.h"

CommentMap::CommentMap( SymMap * symbols ) {
   load( "std/comments.std", 0, symbols );
}

CommentMap::~CommentMap(  ) {
   map < u32, char *>::const_iterator itr = commentmap.begin(  );
   while( itr != commentmap.end(  ) ) {
     delete[] itr->second;
      ++itr;
   }
   itr = refcommentmap.begin(  );
   while( itr != refcommentmap.end(  ) ) {
      delete[] itr->second;
      ++itr;
   }
}

int CommentMap::load( const char *commentfile, u32 baseaddr, SymMap * symbols ) {
   if( !commentfile )
      return 0;
   ifstream data( commentfile, ios::in );
   if( !data ) {
      console.output( "Could not open comment file: %s\n", commentfile );
      return -1;
   }
   console.output( "Loading comments from %s: ", commentfile );
   char *buffer = new char[BUFLEN];
   char *symbol = new char[BUFLEN];
   u32 addr;
   // Symbol *must* exist in SymMap prior to using symbol referencing
   // formats the first space on a line is the separator between the
   // symbol +/- offset and the Comment text, extranious spaces will
   // result in error. If the last character before the space is
   // a '*', then the comment will be displayed whenever the symbol
   // is referenced. Multiple comments for the same symbol/address
   // will cause the previous comment to be overwritten. Comments can
   // be specified for both the address of and for references to a
   // location. Usage of the '*' flag is ignored for offset symbol
   // referenced comments.
   // examples: 0x000A3E Comment at absolute address
   // 0x000A3E* Comment at references to an absolute address
   // blah+0x2 Comment at positive two offset of symbol
   // blah-0x2 Comment at negative two offset of symbol
   // blah* Comment displayed at references to a symbol
   // blah+0x2 Comment at positive offset of symbol
   // blah-0x2 Comment at negative offset of symbol
   data.getline( buffer, BUFLEN );
   int counter = 0;
   while( !data.eof(  ) ) {
      bool atref = false;
      char *temp = strchr( buffer, ' ' );
      if( temp != NULL )
         atref = ( *( temp - 1 ) == '*' );   // has '*' flag?
      if( atref )
	memmove( temp - 1, temp, strlen(temp)+1 );
      if( strncmp( "0x", buffer, 2 ) == 0 ) {   // format "0xXXXXXX Comment\n"
         sscanf( buffer, "0x%X", &addr );
         add( addr + baseaddr, strchr( buffer, ' ' ) + 1, atref );
         ++counter;
      }
      else if( sscanf( buffer, "%s", symbol ) == 1 && symbols->addrof( symbol ) != 0xFFFFFFF0 ) {
         if( strchr( symbol, '+' ) != NULL ) {  // format "Symbol+0xXX
            // Comment\n"
            char *plus = strchr( buffer, '+' );
            strncpy( symbol, buffer, plus - buffer );
            symbol[plus - buffer] = '\0';
            sscanf( plus + 1, "0x%x", &addr );
            add( symbols->addrof( symbol ) + addr, strchr( buffer, ' ' ) + 1 );
            ++counter;
         }
         else if( strchr( symbol, '-' ) != NULL ) {   // format "Symbol-0xXX
            // Comment\n"
            char *minus = strchr( buffer, '-' );
            strncpy( symbol, buffer, minus - buffer );
            symbol[minus - buffer] = '\0';
            sscanf( minus + 1, "0x%x", &addr );
            add( symbols->addrof( symbol ) - addr, strchr( buffer, ' ' ) + 1 );
            ++counter;
         }
         else {                 // format "Symbol Comment\n"
            add( symbols->addrof( symbol ), strchr( buffer, ' ' ) + 1, atref );
            ++counter;
         }
      }
      data.getline( buffer, BUFLEN );
   }
   data.close(  );
   console.output( "%d comments.\n", counter );
   delete[] buffer;
   delete[] symbol;
   return 0;
}

const char *CommentMap::get( u32 addr, bool atref ) {
   if( !atref ) {
      if( commentmap.find( addr ) != commentmap.end(  ) )
         return commentmap[addr];
   }
   else if( refcommentmap.find( addr ) != refcommentmap.end(  ) )
      return refcommentmap[addr];
   return NULL;
}

void CommentMap::add( u32 addr, const char *comment, bool atref ) {
   char *temp = new char[strlen( comment ) + 1];
   strcpy( temp, comment );
   if( !atref ) {
      if( commentmap.find( addr ) != commentmap.end(  ) ) {
         delete commentmap[addr];
      }
      commentmap[addr] = temp;
   }
   else {
      if( refcommentmap.find( addr ) != refcommentmap.end(  ) ) {
         delete refcommentmap[addr];
      }
      refcommentmap[addr] = temp;
   }
}

// automatically fills the commentmap with entries like: 0x123 "This is a String"
// restricted to strings of at least MIN_STRING_LENGTH characters and NULL terminated
void CommentMap::findStrings( const u8 * firmware, u32 length, u32 baseaddr ) {
   console.output( "finding strings in firmware: " );
   u32 start = 0;
   int counter = 0;
   for( u32 i = 0; i < length; ++i ) {
     char c = firmware[i];
      if( c == 0 ) {            // got a string
         int strlength = i - start;
         if( strlength >= MIN_STRING_LENGTH && strlength < ( BUFLEN - 4 ) ) {
            char buffer[BUFLEN] = "";
            sprintf( buffer, "\"%s\"", firmware + start );
            add( baseaddr + start, buffer, true );
            ++counter;
         }
      }
      if( !isprint( c ) )
         start = i + 1;
   }
   console.output( "%d strings found\n", counter );
}

const int CommentMap::dump( const char *file ) {
   ofstream data( file, ios::out );
   if( !data ) {
      console.output( "Could not open comment file: %s\n", file );
      return -1;
   }
   char *buffer = new char[128 * 1024];
   map < u32, char *>::const_iterator itr = commentmap.begin(  );
   while( itr != commentmap.end(  ) ) {
      int length = sprintf( buffer, "0x%06X %s\n", itr->first, itr->second );
      data.write( buffer, length );
      ++itr;
   }
   itr = refcommentmap.begin(  );
   while( itr != refcommentmap.end(  ) ) {
      int length = sprintf( buffer, "0x%06X %s\n", itr->first, itr->second );
      data.write( buffer, length );
      ++itr;
   }
   delete[] buffer;
   data.close(  );
   return 0;
}
