/*  Project: Open Source Gmini Tools
    file: disasm.cpp, main procedure for gmdisasm

    Copyright (C) 2004  The gmEmu Project <gmemu@donat.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <iostream>
#include <fstream>
#include <stdarg.h>
#include <stdio.h>
//#ifndef GNU_LIBC
//#include "getopt.h" // yanked from GNU libc source

#include "types.h"
#include "util.h"
#include "decodeOP.h"
#include "formatOP.h"
#include "SymMap.h"
#include "CommentMap.h"
#include "Console.h"

#define DISASMNAME "disasm"
#define HTML html

bool html = true;

void listwrite( ofstream & data, const char *format, ... ) {
   char buffer[128 * 1024];
   va_list vargs;
   va_start( vargs, format );
   vsprintf( buffer, format, vargs );
   va_end( vargs );
   data.write( buffer, strlen( buffer ) );
}
struct memory {
   u32 ip;
   u32 max;
   u8* data;
} flash;
SymMap *symbols;
CommentMap *comments;
map<u32,u32> loadmap;

const bool isLoaded( u32 addr ) {
   map<u32,u32>::iterator itr = loadmap.begin();
   while( itr != loadmap.end() ) {
      if( addr >= itr->first && addr < itr->second )
         return true;
      ++itr;
   }
   return false;
}

void setLoad( u32 start, u32 length ) {
   u32 end = start + length;
   bool done;
   do {
      done = true;
      map<u32,u32>::iterator itr = loadmap.begin();
      while( itr != loadmap.end() ) {
         bool lap1 = false;
         bool lap2 = false;
         if( start < itr->first && end > itr->first )
            lap1 = true;
         if( start < itr->second && end > itr->second )
            lap2 = true;
         if( lap1 || lap2 ) {
            if( lap1 && !lap2 )
               end = itr->second;
            if( !lap1 && lap2 )
               start = itr->first;
            loadmap.erase( itr->first );
            done = false;
            break;
         }
         else if( start > itr->first && end < itr->second ) {
            length = 0;
            break;
         }
         ++itr;
      }
   } while( !done );
   if( length != 0 )
      loadmap[ start ] = end;
}

bool aazload( const char* filename ) {
   u8 *bin;
   int size = loadFile( filename, &bin );
   if( size <= 0 )
      return false;
   console.output( "File: \"%s\" (%x)\n", filename, size );
   // Parse AAZ header
   u32 header_version = bu32( bin + 0 );
   if( header_version != 2 ) {
      console.output( "  Unknown header version: %d\n", header_version );
      delete bin;
      return false;
   }
   u32 loadaddress = bu32( bin + 4 );
   console.output( "  Loading to: 0x%x\n", loadaddress );
   u32 flash_size = bu32( bin + 8 );
   console.output( "  Flash size: 0x%x\n", flash_size );
   if (flash_size != (unsigned) size - 0x100) {
      console.output( "  Expected size: 0x%x, aborting.\n", size - 0x100 );
      delete bin;
      return false;
   }
   char version[5];
   cstr( version, bin + 12, 4 );
   console.output( "  Version   : \"%s\"\n", version );
   u16 chksum = bu16( bin + 16 );
   console.output( "  Checksum16: 0x%04x", chksum );
   u16 chkchksum = chksum32( bin + 0x100, flash_size ) & 0xffff;
   if( chksum == chkchksum )
      console.output( " ok!\n" );
   else
      console.output( " wrong! 0x%04x\n", chkchksum );
   u16 flashHeader = bu16( bin + 18 );
   console.output( "  Flash header: %s\n", flashHeader ? "true" : "false" );
   // -> flashHeader
   char description[13];
   cstr( description, bin + 20, 12 );
   console.output( "  Magic     : \"%s\"\n", description );
   u32 resetaddr = bu32( bin + 32 );
   console.output( "  ResetAddr : 0x%x\n", resetaddr );
   // -> where execution begins
   u32 unknown = bu32( bin + 36 );
   console.output( "  unknown   : 0x%x\n", unknown );
   if( chksum != chkchksum ) {
      console.output( "  Bad AAZ, load aborting.\n" );
      return false;
   }
   memcpy( flash.data + loadaddress, bin + ( flashHeader ? 0 : 0x100 ) , size - ( flashHeader ? 0 : 0x100 ) );
   delete bin;
   if( flash.ip > loadaddress )
      flash.ip = loadaddress;
   if( flash.max < ( loadaddress + size - ( flashHeader ? 0 : 0x100 ) ) )
      flash.max = ( loadaddress + size - ( flashHeader ? 0 : 0x100 ) );
   if( flashHeader )
      symbols->setData( loadaddress, 0x100 );
   setLoad( loadaddress, size - ( flashHeader ? 0 : 0x100 ) );
   return true;
}

void disasmTCC730() {
   u8 *end = flash.data + flash.max;
   u32 ipsave = flash.ip;
   console.output( "First pass\n" );
   while( ( flash.data + flash.ip ) < end ) {
      if( isLoaded( flash.ip ) && !symbols->isData( flash.ip ) ) {
         u16 op1 = bu16( flash.data + flash.ip );
         u16 op2 = ( ( flash.data + flash.ip + 2 ) < end ) ? bu16( flash.data + flash.ip + 2 ) : 0;
         int mnnum = decodeOP( op1 );
         formatOP( symbols, comments, mnnum, op1, op2, flash.ip + opsize( mnnum ) );
         flash.ip += opsize( mnnum );
      }
      else 
         flash.ip += 2;
   }
   flash.ip = ipsave;
   char *listfile = new char[ ( HTML ? 14 : 13 ) ];
   sprintf( listfile, "./disasm.%s", ( HTML ? "html" : "lst" ) );
   console.output( "Second pass, out to %s\n", listfile );
   ofstream data( listfile, ios::out );
   if( !data ) {
      console.output( "Could not open list file: %s\n", listfile );
      return;
   }
   char *refbuf = new char[128 * 1024];
   if( HTML )
      listwrite( data, "<html><head><title>%s</title></head><body><pre>\n", listfile );
   bool firstgap = false;
   while( ( flash.data + flash.ip ) < end ) {
      if( isLoaded( flash.ip ) && !symbols->isData( flash.ip ) ) {
         u16 op1 = bu16( flash.data + flash.ip );
         u16 op2 = ( ( flash.data + flash.ip + 2 ) < end ) ? bu16( flash.data + flash.ip + 2 ) : 0;
         int mnnum = decodeOP( op1 );
         char *inst = formatOP( symbols, comments, mnnum, op1, op2, flash.ip + opsize( mnnum ), HTML );
         const char *sym = symbols->get( flash.ip );
         const char *comment = comments->get( flash.ip );
         if( sym != NULL && strncmp( "data_", sym, 5 ) != 0 ) {
            if( HTML )
               listwrite( data, "<a name=\"%s\">%s</a>:", sym, quote( sym ) );
            else
               listwrite( data, "%s:", sym );
            if( sym[0] == 'r' && sym[1] == 'e' && sym[2] == 'l' && sym[3] == '_' )
               if( symbols->getref( flash.ip, refbuf, HTML ) != NULL && strlen( refbuf ) > 4 )
                  listwrite( data, "%s", refbuf );
            if( comment != NULL )
               listwrite( data, "                                        ; %s", comment );
            if( sym[0] != 'r' || sym[1] != 'e' || sym[2] != 'l' || sym[3] != '_' )
               if( symbols->getref( flash.ip, refbuf, HTML ) != NULL )
                  listwrite( data, "\n%s", refbuf );
            listwrite( data, "\n" );
         }
         char str_op1[5];
         char str_op2[6];
         char str_mnnum[5];
         sprintf( str_op1, "%04X", op1 );
         sprintf( str_op2, " %04X", op2 );
         if( mnnum < 164 )
            sprintf( str_mnnum, " %03u", mnnum );
         else
            sprintf( str_mnnum, "E%03u", mnnum - 164 );
         if( HTML && ( mnnum == 117 || mnnum == 118 || mnnum == 119 ||
                       mnnum == 120 || mnnum == 136 || mnnum == 137 ||
                       mnnum == 150 || mnnum == 151 || mnnum == 153 ||
                       mnnum == 154 || mnnum == 156 || mnnum == 157 ||
                       mnnum == 158 || mnnum == 159 || mnnum == 160 || mnnum == 161 ) )
            listwrite( data, "<a name=\"%06X\">%06X</a>:", flash.ip, flash.ip );
         else
            listwrite( data, "%06X:", flash.ip );
         listwrite( data, "(%s) %s%s: %s", str_mnnum, str_op1, ( ( opsize( mnnum ) == 4 ) ? str_op2 : "     " ), inst );

         if( sym == NULL && comment != NULL )
            listwrite( data, "    ; %s\n", comment );
         else
            listwrite( data, "\n" );
         if( HTML && op1 == 0x87E6 )
            listwrite( data, "<hr>\n" );
         flash.ip += opsize( mnnum );
         firstgap = true;
      }
      else {
         if( HTML && firstgap ) {
            listwrite( data, "<hr size=\"8\">\n" );
            firstgap = false;
         }
         flash.ip += 2;
      }
   }
   if( HTML )
      listwrite( data, "</pre></body></html>\n" );
   data.close(  );
   delete[] listfile;
}


int main( int argc, char **argv ) {
   if (argc > 1 && !strcmp(argv[1], "-l")) {
      //horrid way of parsing options. Bah.
      html = false;
      argc--;
      argv++;
   }
//**************************
   argc--;
   argv++;
   // init global data
   flash.ip = 0xFFFFFFFF;
   flash.max = 0;
   flash.data = new u8[ 0x400000 ];
   symbols = new SymMap(  );
   comments = new CommentMap( symbols );

   // parse args
   for (int i = 0; i<argc; i++) {
      const char* fn = argv[i];
      const char* ext = fileextension(fn);

      if (strcmp(ext, ".aaz") == 0) {
         if( !aazload( fn ) ) {
            console.output( "Failed to load firmware file: %s.\n", fn );
            abort(  );
         }
      } else if (strcmp(ext, ".map") == 0) {
       symbols->load( fn );
      } else {
         console.output( "Unrecognized file: %s\n", fn );
         abort();
      }

      delete[] ext;
   }

   // Start disassembly?
   if( flash.ip != 0xFFFFFFFF ) {
      map<u32,u32>::iterator itr = loadmap.begin();
      while( itr != loadmap.end() ) {
         comments->findStrings( flash.data + itr->first, itr->second - itr->first, itr->first );
         ++itr;
      }
      disasmTCC730();
      symbols->dump( "./disasm.map" );
      comments->dump( "./disasm.tag" );
   }

   delete flash.data;
   delete symbols;
   delete comments;
   return 0;
}
